
import numpy as np
import re
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import LabelEncoder
# Create a simple parametrable chord prediction algorithm


class WindowedPredictor:
    """Simple next word predictor using a context window
    """
    _vector_size = 10
    _window = 3
    _memory = 4

    def __init__(self, clf, vector_size=_vector_size, window=_window, memory=_memory, **config):
        """

        Parameters
        ----------

        clf:
        vector_size: int
                Size of Word2Vec embedding (default=10)
        window: int
                Window size used for Word2Vec, (default=3)
        memory: int
                How much tokens are used to predict the next one (default)
        config: Any
                Other parameters
        """
        self.clf = clf
        self.wv = None
        self.config = config
        self.vector_size = vector_size
        self.memory = memory
        self.window = window
        self.encoder = LabelEncoder()


    def save(self, filepath):
        """

        Parameters
        ----------
        filepath :
            

        Returns
        -------

        """
        import joblib
        joblib.dump(self, filepath)

    @classmethod
    def load(cls, filepath):
        """

        Parameters
        ----------
        filepath : str
                   Filepath of the model
            

        Returns
        -------

        """
        import joblib
        return joblib.load(filepath)

    def cross_val_score(self, data, **kwargs):
        """Return the cross val scores of this model using the data provided

        Parameters
        ----------
        data :
            Input data
        **kwargs :
            arguments that will be passed to sklearn.model_selection.cross_val_score
            

        Returns
        -------

        """
        from sklearn.model_selection import cross_val_score
        from sklearn.metrics import make_scorer
        X, Y = self.transform(data)
        y = self.encoder.fit_transform(Y)
        scorer = 'accuracy'
        return cross_val_score(self.clf, X, y, scoring=scorer, **kwargs)

    def eval(self, data):
        from sklearn.metrics import accuracy_score
        X, Y = self.transform(data, train=False)
        y_true = Y
        y_pred = self.encoder.inverse_transform(self.clf.predict(X))
        return accuracy_score(y_true, y_pred)


    def filter(self, data):
        """Filter some data by default before training, it can help to reduce the vocab size

        Parameters
        ----------
        data :
            return:

        Returns
        -------

        """
        return data

    def train_embeding(self, data):
        """

        Parameters
        ----------
        data :
            

        Returns
        -------

        """
        data = self._pad_data(data)
        self._train_embedding(data)
        data_transformed = self._embed(data)
        return data_transformed, data


    def _train_embedding(self, data):
        try:
            import gensim
        except ImportError:
            raise ImportError('Please install gensim to use this module')
        model = gensim.models.Word2Vec(data, min_count=1, vector_size=self.vector_size, window=self.window)
        self.wv = model.wv

    def _pad_data(self, data):
        # Pad with START TOKEN
        data = [['START'] * self.memory + self.filter(d) for d in data]
        return data

    def _embed(self, data):
        data_transformed = [[self.wv.vectors[self.wv.get_index(chord, default=self.wv.get_index('START'))] for chord in song] for song in data]
        return data_transformed

    def embed(self, data):
        """

        Parameters
        ----------
        data

        Returns
        -------

        """

        data = self._pad_data(data)
        data_transformed = self._embed(data)
        return data_transformed, data

    def fit(self, data):
        """

        Parameters
        ----------
        data :
            

        Returns
        -------

        """
        X, Y = self.transform(data)
        y = self.encoder.fit_transform(Y)
        self.clf.fit(X, y)
        return self

    def transform(self, data, train=True):
        """

        Parameters
        ----------
        data :
            

        Returns
        -------

        """
        from sklearn.utils import shuffle
        if train:
            data_transformed, filtered_data = self.train_embeding(data)
        else:
            data_transformed, filtered_data = self.embed(data)
        # Create X, Y dataset
        X = [np.asarray(song[chord_idx:chord_idx + self.memory]).ravel() for song in data_transformed
             for chord_idx, chord in enumerate(song[self.memory:])
             ]

        Y = [song[chord_idx + self.memory] for song in filtered_data for chord_idx, chord in
             enumerate(song[self.memory:])]
        X = np.asarray(X)
        Y = np.asarray(Y)
        X, Y = shuffle(X, Y)
        return X, Y

    def predict(self, sequence, topn=None):
        """

        Parameters
        ----------
        sequence : List[str]
            Sequence of previous tokens on which to predict the next token
        topn : int | None, optional
            If topn is None, then returns the highest score prediction
            Otherwise returns the list of top-n ranked predicted tokens (highest first) (Default value = None)

        Returns
        -------
        token_result : str or List[str]
            

        Examples
        --------
        Simple usage to predict next chord :

        >>> from musiclang.predict.predictors import WindowedPredictor
        >>> from musiclang.library import *
        >>> from musiclang.predict.tokenizers import ChordTokenizer
        >>> tokenizer = ChordTokenizer()
        >>> predictor = WindowedPredictor.load("ChordPredictor")
        >>> chord_progression = (I % I.M) + (VI['6'] % I.M) + (II['2'] % I.M)
        >>> chord_tokens = tokenizer.tokenize(chord_progression)
        >>> predictor.predict(chord_tokens)
        '(V['65'] % I.M)'

        To predict the three best chords :

        >>> predictor.predict(chord_tokens, topn=3)
        ['(V['65'] % I.M)', '(V['7'] % I.M)', '(I['64'] % I.M)']

        """
        import numpy as np
        if len(sequence) < self.memory:
            to_fill = self.memory - len(sequence)
            sequence = ['START'] * to_fill + sequence

        data = np.asarray([self.wv[s] for s in sequence[-self.memory:]]).ravel()
        # Classification
        prediction = self.clf.predict_proba([data])[0]
        if topn is None:
            prediction = np.argmax(prediction)
            token_result = self.encoder.inverse_transform([prediction])[0]
            return token_result
        elif topn > 0:
            predictions = np.argsort(-prediction)
            token_result = self.encoder.inverse_transform(predictions[:topn]).tolist()
            return token_result
        else:
            raise ValueError('Invalid value for topn, it should be None or a positive integer')



class MelodyPredictor(WindowedPredictor):
    """ """

    def filter(self, data):
        """

        Parameters
        ----------
        data :
            

        Returns
        -------

        """
        return [d for d in data if 'augment' not in d]


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

INSTRUMENTS_DICT = {
    "piano": 0,
    "pizzicato": 45,
    "bright_piano": 1,
    "electric_piano": 2,
    "honky_tonk": 3,
    "electric_piano_1": 4,
    "electric_piano_2": 5,
    "harpsichord": 6,
    "clavi": 7,
    "celesta": 8,
    "glockenspiel": 9,
    "music_box": 10,
    "vibraphone": 11,
    "marimba": 12,
    "xylophone": 13,
    "tubular_bells": 14,
    "dulcimer": 15,
    "drawbar_organ": 16,
    "percussive_organ": 17,
    "rock_organ": 18,
    "church_organ": 19,
    "reed_organ": 20,
    "accordion": 21,
    "harmonica": 22,
    "tango_accordion": 23,
    "acoustic_guitar": 24,
    "steel_guitar": 25,
    "jazz_guitar": 26,
    "clean_guitar": 27,
    "muted_guitar": 28,
    "overdriven_guitar": 29,
    "distortion_guitar": 30,
    "harmonic_guitar": 31,
    "acoustic_bass": 32,
    "electric_bass_finger": 33,
    "electric_bass_pick": 34,
    "fretless_bass": 35,
    "slap_bass_1": 36,
    "slap_bass_2": 37,
    "synth_bass_1": 38,
    "synth_bass_2": 39,
    "violin": 40,
    "viola": 41,
    "cello": 42,
    "contrabass": 43,
    "tremolo_string": 44,
    "harp": 46,
    "timpani": 47,
    "string_ensemble_1": 48,
    "string_ensemble_2": 49,
    "synth_string_1": 50,
    "synth_string_2": 51,
    "choir_aahs": 52,
    "choir_oohs": 53,
    "synth_choir": 54,
    "orchestra_hit": 55,
    "trumpet": 56,
    "trombone": 57,
    "tuba": 58,
    "muted_trumpet": 59,
    "french_horn": 60,
    "brass_section": 61,
    "synth_brass_1": 62,
    "synth_brass_2": 63,
    "soprano_sax": 64,
    "alto_sax": 65,
    "tenor_sax": 66,
    "baritone_sax": 67,
    "oboe": 68,
    "english_horn": 69,
    "bassoon": 70,
    "clarinet": 71,
    "piccolo": 72,
    "flute": 73,
    "recorder": 74,
    "pan_flute": 75,
    "blown_bottle": 76,
    "shakuhachi": 77,
    "whistle": 78,
    "ocarina": 79,
    "square_lead": 80,
    "sawtooth_lead": 81,
    "calliope_lead": 82,
    "chiff_lead": 83,
    "charang_lead": 84,
    "voice_lead": 85,
    "fifths_lead": 86,
    "bass_lead": 87,
    "pad_new_age": 88,
    "pad_warm": 89,
    "pad_polysynth": 90,
    "pad_choir": 91,
    "pad_bowed": 92,
    "pad_metallic": 93,
    "pad_halo": 94,
    "pad_sweep": 95,
    "fx_rain": 96,
    "fx_soundtrack": 97,
    "fx_crystal": 98,
    "fx_atmosphere": 99,
    "fx_brightness": 100,
    "fx_globlins": 101,
    "fx_echoes": 102,
    "fx_sci_fi": 103,
    "sitar": 104,
    "banjo": 105,
    "shamisen": 106,
    "koto": 107,
    "kalimba": 108,
    "bagpipe": 109,
    "fiddle": 110,
    "shanai": 111,
    "tinkle_bell": 112,
    "agogo": 113,
    "steel_drums": 114,
    "woodblock": 115,
    "taiko_drum": 116,
    "melodic_tom": 117,
    "synth_drum": 118,
    "reverse_cymbal": 119,
    "guitar_fret_noise": 120,
    "breath_noise": 121,
    "seashore": 122,
    "bird_tweet": 123,
    "telephone_ring": 124,
    "helicopter": 125,
    "applause": 126,
    "gunshot": 127
}

REVERSE_INSTRUMENT_DICT = {
    val: key for key, val in INSTRUMENTS_DICT.items()
}

OCTAVES = {
    INSTRUMENTS_DICT['piccolo']: 1,
    INSTRUMENTS_DICT['flute']: 1,
    INSTRUMENTS_DICT['oboe']: 0,
    INSTRUMENTS_DICT['english_horn']: 0,
    INSTRUMENTS_DICT['clarinet']: 0,
    INSTRUMENTS_DICT['bassoon']: -1,
    INSTRUMENTS_DICT['french_horn']: -1,
    INSTRUMENTS_DICT['trumpet']: -1,
    INSTRUMENTS_DICT['trombone']: -2,
    INSTRUMENTS_DICT['tuba']: -2,
    INSTRUMENTS_DICT['piano']: -1,
    INSTRUMENTS_DICT['violin']: 0,
    INSTRUMENTS_DICT['viola']: 0,
    INSTRUMENTS_DICT['cello']: -2,
    INSTRUMENTS_DICT['contrabass']: -2,
    INSTRUMENTS_DICT['harp']: -1,
    INSTRUMENTS_DICT['acoustic_guitar']: -1,
    INSTRUMENTS_DICT['timpani']: -2
}

OCTAVES = {key: 0 for key in OCTAVES.keys()}


from .composing import create_counterpoint, create_counterpoint_on_score
from .melody import CircularPermutationMelody, SelectRangeMelody, ReverseMelody, \
    InvertMelody, ReverseMelodyWithoutRhythm
from .note import LimitRegister, ApplySilence, TransposeDiatonic, TransposeChromatic, ApplyContinuation
from .score import RepeatScore
from .features import ExtractMainTonality, MostCommonTonalities
from .chord import RepeatChord, Modulate, ModulateKeepOriginMode
from .score_merger import ConcatScores, ConcatWithPattern, TakeSlice, TakeFirst, TakeLast, TakeIdx
from .composing import VoiceLeading, Patternator
from .dynamics import PitchDynamizer
from .melody import ContinuationWhenSameNote

__all__ = ['create_counterpoint', 'create_counterpoint_on_score',
           'CircularPermutationMelody', 'SelectRangeMelody', 'ReverseMelody', 'InvertMelody',
           'RepeatScore', 'ReverseMelodyWithoutRhythm',
            'LimitRegister', 'ApplySilence', 'TransposeDiatonic', 'TransposeChromatic', 'ApplyContinuation',
            'ExtractMainTonality', 'MostCommonTonalities',
            'RepeatChord', 'Modulate', 'ModulateKeepOriginMode',
           'ConcatScores', 'ConcatWithPattern', 'TakeSlice', 'TakeFirst', 'TakeLast', 'TakeIdx',
           'VoiceLeading', 'PitchDynamizer', 'ContinuationWhenSameNote', 'Patternator'
           ]



import numpy as np
from itertools import product
from musiclang import Chord, Tonality

TEMPLATES = [
                ('', [0, 4, 7], 1),
                ('(m3)', [0, 3, 7], 1),
                ('(m3)(b5)', [0, 3, 6], 1.0),
                ('(m3)(b5)[add6]', [0, 3, 6, 9], 1.0),
                ('(+)', [0, 4, 8], 0.8),
                ('(sus2)', [0, 2, 7], 0.9),
                ('(sus4)', [0, 5, 7], 0.9),
                ('(M7)', [0, 4, 7, 11], 0.7),
                ('(m3)(m7)', [0, 3, 7, 10], 0.7),
                ('(m3)(M7)', [0, 3, 7, 10], 0.7),
                ('(m7)', [0, 4, 7, 10], 1.0),
                ('(m3)(m7)(b5)', [0, 3, 6, 10], 0.95),
        ]

EXTENSIONS = [  '',
                '7',
                '(+)',
                '(sus2)',
                '(sus4)',
               ]
CHORD_TYPE_TO_PITCHES = {t[0]: t[1] for t in TEMPLATES}
CHORD_TYPES = np.asarray([t[0] for t in TEMPLATES])
PITCH_CLASS = np.arange(12)

CHROMA_VECTORS = [t[1] for t in TEMPLATES]
COEFFS = np.asarray([t[2] for t in TEMPLATES])
# Convert chroma_vectors into a 12xn_chords matrix with each row being the indexes of the chroma_vector
CHROMA_VECTORS_MATRIX = np.asarray([[(1.0*(i%12)) in chroma_vector for i in range(12)] for chroma_vector in CHROMA_VECTORS])
CHROMA_VECTORS_MATRIX = np.concatenate([np.roll(CHROMA_VECTORS_MATRIX, i, axis=1) for i in range(12)], axis=0)
COEFFS = np.concatenate([COEFFS for i in range(12)], axis=0)

MODES = ['M', 'm']
TONALITIES = list(range(12))
ROMAN_NUMERALS = list(range(7))

PITCHES_DICT = {(roman, tonality_root, tonality_mode, extension): Chord(roman, tonality=Tonality(tonality_root, mode=tonality_mode))[extension].chord_extension_pitch_classes
                for roman, tonality_root, tonality_mode, extension in product(ROMAN_NUMERALS, TONALITIES, MODES, EXTENSIONS)}

# Remove all augmented with minor thirds
PITCHES_DICT = {k: v for k, v in PITCHES_DICT.items() if not (k[3] == '(+)' and ((v[1] - v[0]) % 12) == 3)}
# Remove the 2, _, m, sus4
PITCHES_DICT = {k: v for k, v in PITCHES_DICT.items() if not (k[3] == '(sus4)' and k[0] == 2 and k[2] == 'm')}
# Remove the 5, _, m, sus2
PITCHES_DICT = {k: v for k, v in PITCHES_DICT.items() if not (k[3] == '(sus2)' and k[0] == 5 and k[2] == 'm')}


# Invert pitches dict to get a dict of pitch classes to potential chord types
PITCH_CLASSES_DICT = {}
for k, v in PITCHES_DICT.items():
    pitch_class = tuple(sorted(v))
    if pitch_class not in PITCH_CLASSES_DICT:
        PITCH_CLASSES_DICT[pitch_class] = []
    PITCH_CLASSES_DICT[pitch_class].append(k)


from .element import Formatter
from musiclang import Chord, Tonality
from musiclang.analyze.roman_parser import analyze_one_chord

class ChordElement(Formatter):

    def __init__(self, text):
        self.text = text

    def to_text(self):
        return ' ' + str(self)

    def __repr__(self):
        return f'{self.text}'

    @classmethod
    def get_key_mode(cls, tonality):
        tab = 'CDEFGAB'
        notes = [0, 2, 4, 5, 7, 9, 11]
        note = tonality.replace(':', '').replace('#', '').replace('b', '').replace('-', '')
        tone = notes[tab.index(note.upper())]
        mode = 'M' if note.upper() == note else 'm'

        tone += tonality.count('#')
        tone -= tonality.count('b')
        tone -= tonality.count('-')

        return tone, mode

    @classmethod
    def get_tonic_chord(cls, tonality):
        key, mode = cls.get_key_mode(tonality)
        if mode == 'm':
            return (ChordElement('i')).parse_with_tonality(tonality)
        elif mode == 'M':
            return (ChordElement('I')).parse_with_tonality(tonality)

    def parse_with_tonality(self, tonality):
        key, mode = self.get_key_mode(tonality)
        final_degree, extension, final_key, final_mode = analyze_one_chord(self.text, key, mode)
        chord = Chord(final_degree,
                      tonality=Tonality(final_key, final_mode))[extension]

        return chord

    def has_pitch(self, pitch, tonality):
        chord = self.parse_with_tonality(tonality)
        return chord.has_pitch(pitch)


from musiclang.write.pitches.pitches_utils import *
from musiclang import Note

def test_relative_scale_up_value():

    new_pitch = relative_scale_up_value(1, 12, [0, 2, 4, 5, 7, 9, 11])

    assert new_pitch == 14

def test_relative_scale_up_value_negative():
    new_pitch = relative_scale_up_value(1, -2, [0, 2, 4, 5, 7, 9, 11])

    assert new_pitch == -1


def test_relative_scale_up_value_octave():

    new_pitch = relative_scale_up_value(8, 12, [0, 2, 4, 5, 7, 9, 11])
    assert new_pitch == 26

def test_relative_scale_up_value_not_in_scale():

    new_pitch = relative_scale_up_value(1, 13, [0, 2, 4, 5, 7, 9, 11])

    assert new_pitch == 14



def test_relative_scale_down_value():
    new_pitch = relative_scale_down_value(1, 12, [0, 2, 4, 5, 7, 9, 11])

    assert new_pitch == 11



def test_relative_scale_down_value_not_in_scale():

    new_pitch = relative_scale_down_value(1, 13, [0, 2, 4, 5, 7, 9, 11])

    assert new_pitch == 12


def test_get_relative_scale_value():

    note = Note("su", 1, 1, 1)
    last_pitch = -2
    scale_pitches = [0, 2, 4, 5, 7, 9, 11]
    result = get_relative_scale_value(note, last_pitch, scale_pitches)

    assert result == 11

def test_get_relative_scale_value_down():

    note = Note("sd", 1, 0, 1)
    last_pitch = -1
    scale_pitches = [0, 2, 4, 5, 7, 9, 11]
    result = get_relative_scale_value(note, last_pitch, scale_pitches)

    assert result == -3


from musiclang.library import *
from musiclang import Score
from musiclang.analyze.score_formatter import ScoreFormatter

expected_text = """
(I % I.m)(
	piano__0=b0.h, 
	piano__1=b1.h, 
	piano__2=b2.h, 
	piano__3=b3.h)+ 
(V['7'] % V.M)(
	piano__0=b0.hd, 
	piano__1=b1.hd, 
	piano__2=b2.hd, 
	piano__3=b3.hd)+ 
(I % I.M)(
	piano__0=b0.hd, 
	piano__1=b1.hd, 
	piano__2=b2.hd, 
	piano__3=b3.hd)+ 
(V['7'] % IV.s.mm)(
	piano__0=b0.w, 
	piano__1=b1.w, 
	piano__2=b2.w, 
	piano__3=b3.w)+ 
(I['64'] % I.m)(
	piano__0=b0.h, 
	piano__1=b1.h, 
	piano__2=b2.h, 
	piano__3=b3.h)+ 
(V['7(sus2)'] % II.M)(
	piano__0=b0.h, 
	piano__1=b1.h, 
	piano__2=b2.h, 
	piano__3=b3.h)
"""


def test_score_formatter_simple():
    text = """
    Time Signature: 4/4
    
    m0 c: i b3 II7
    m1 b2 I
    m2 N7
    m3 Cad64 b3 D: V7[sus2]
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text)
    assert score == expected_score


expected_text2 = """
(I % I.M)(
	piano__0=b0, 
	piano__1=b1, 
	piano__2=b2, 
	piano__3=b3)+ 
(V['7'] % I.M)(
	piano__0=b0.h, 
	piano__1=b1.h, 
	piano__2=b2.h, 
	piano__3=b3.h) 
"""


def test_score_formatter_time_signature_68():
    text = """
    Time Signature: 6/8
    m0 C: I b1.66 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text2)
    assert score == expected_score

expected_text2b = """
(I % I.M)(
	piano__0=b0, 
	piano__1=b1, 
	piano__2=b2, 
	piano__3=b3)+ 
(V['7'] % I.M)(
	piano__0=b0.h, 
	piano__1=b1.h, 
	piano__2=b2.h, 
	piano__3=b3.h) 
"""


def test_score_formatter_start_measure1():
    text = """
    Time Signature: 6/8
    m1 C: I b1.66 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text2b)
    assert score == expected_score


expected_text3 = """
(I % I.M)(
	piano__0=b0, 
	piano__1=b1, 
	piano__2=b2, 
	piano__3=b3)+ 
(V['7'] % I.M)(
	piano__0=b0.hd, 
	piano__1=b1.hd, 
	piano__2=b2.hd, 
	piano__3=b3.hd) 
"""


def test_score_formatter_time_signature_22():
    text = """
    Time Signature: 2/2
    m0 C: I b1.5 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text3)
    assert score == expected_score


expected_text4 = """
(I % I.M)(
	piano__0=b0.w, 
	piano__1=b1.w, 
	piano__2=b2.w, 
	piano__3=b3.w)+ 
(V['7'] % I.M)(
	piano__0=b0.hd, 
	piano__1=b1.hd, 
	piano__2=b2.hd, 
	piano__3=b3.hd) 
"""


def test_score_formatter_time_signature_change():
    text = """
    Time Signature: 2/2
    m0 C: I
    Time Signature: 3/4
    m1 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse(allow_multi_signature=True)
    expected_score = Score.from_str(expected_text4)
    assert score == expected_score


expected_text5 = """
(I % I.M)(
	piano__0=b0.wd, 
	piano__1=b1.wd, 
	piano__2=b2.wd, 
	piano__3=b3.wd)+ 
(V['7'] % I.M)(
	piano__0=b0, 
	piano__1=b1, 
	piano__2=b2, 
	piano__3=b3) 
"""


def test_score_formatter_time_signature_change_complex():
    text = """
    Time Signature: 2/2
    m0 C: I
    Time Signature: 3/4
    m1 b3 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse(allow_multi_signature=True)
    expected_score = Score.from_str(expected_text5)
    assert score == expected_score


expected_text6 = """
(I % I.M)(
	piano__0=b0.augment(frac(8, 1)), 
	piano__1=b1.augment(frac(8, 1)), 
	piano__2=b2.augment(frac(8, 1)), 
	piano__3=b3.augment(frac(8, 1)))+ 
(V['7'] % I.M)(
	piano__0=b0.w, 
	piano__1=b1.w, 
	piano__2=b2.w, 
	piano__3=b3.w) 
"""


def test_score_formatter_multi_bars():
    text = """
    Time Signature: 2/2
    m0 C: I
    m2 V7
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text6)
    assert score == expected_score


expected_text7 = """
(I % I.M)(
	piano__0=b0.w, 
	piano__1=b1.w, 
	piano__2=b2.w, 
	piano__3=b3.w)+ 
(I % I.M)(
	piano__0=b0.w, 
	piano__1=b1.w, 
	piano__2=b2.w, 
	piano__3=b3.w)
"""

def test_score_ignore_var():
    text = """
    Time Signature: 4/4
    m0 C: I
    m0var1 c: III
    m0var2 c: III
    m1 I
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text7)
    assert score == expected_score


expected_text8 = """
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w)+
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w)
"""


def test_score_repeated_bar():
    text = """
    Time Signature: 4/4
    m0 C: I
    m1 = m0
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text8)
    assert score == expected_score


expected_text9 = """
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w)+
(V % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(V % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(VI % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) 
	
"""


def test_score_multi_repeat():
    text = """
    Time Signature: 4/4
    m0 C: I
    m1 V
    m2-3 = m0-1
    m4 vi
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text9)
    assert score == expected_score


expected_text10 = """
(I % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(VI % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) 
"""

def test_score_pickup_bar():
    text = """
    Time Signature: 4/4
    m0 C: b3 I
    m1 V
    m2 I
    m3 vi
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text10)
    assert score == expected_score


expected_text11 = """
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w,
	violin__0=r.w
	) +
(V % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w,
	violin__0=r.w
	) +
(I % I.M)(
    piano__0=b0.o(-2).w,
    piano__1=b2.o(-1).w,
    piano__2=b3.w,
    violin__0=b1.o(1).w,
    piano__3=r.w
    )+
(VI % I.M)(
    piano__0=b0.o(-2).w,
    piano__1=b2.o(-1).w,
    piano__2=b3.w,
    violin__0=b1.o(1).w,
    piano__3=r.w
    )
"""
def test_score_change_voicing():
    text = """
    Time Signature: 4/4
    m0 C: I
    m1 V
    !instruments piano    piano    piano violin
    !voicing     b0.o(-2) b2.o(-1) b3    b1.o(1)
    m2 I
    m3 vi
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text11)
    assert score == expected_score


expected_text12 = """
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(V % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w) +
(I % I.M)(
    piano__0=b0.o(-2).w,
    piano__1=b2.o(-1).w,
    piano__2=b3.w,
    violin__0=b1.o(1).w)+
(VI % I.M)(
    piano__0=b0.o(-2).w,
    piano__1=b2.o(-1).w,
    piano__2=b3.w,
    violin__0=b1.o(1).w)
"""
def test_voice_leading_work():
    text = """
    !voice_leading
    Time Signature: 4/4
    m0 C: I
    m1 V
    !instruments piano    piano    piano violin
    !voicing     b0.o(-2) b2.o(-1) b3    b1.o(1)
    m2 I
    m3 vi
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    not_expected_score = Score.from_str(expected_text12)
    assert score != not_expected_score

expected_text13 = """
(I % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.h + b0.h) +
(V % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.w)
"""
def test_rhythm():
    text = """
    Time Signature: 4/4
    !instruments piano violin
    !voicing  b0.o(-2) b0
    !rhythm violin x.x.|x...
    m0 C: I
    m1 V
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text13)
    assert score == expected_score



expected_text14 = """
(I % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.w) +
(V % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.w)+
(I % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.h + b0.h) +
(V % I.M)(
	piano__0=b0.o(-2).w,
	violin__0=b0.w)
"""
def test_rhythm_change():
    text = """
    Time Signature: 4/4
    !instruments piano violin
    !voicing  b0.o(-2) b0
    m0 C: I
    m1 V
    !rhythm violin x.x.|x...
    m2 C: I
    m3 V
    """
    sf = ScoreFormatter(text)
    score = sf.parse()
    expected_score = Score.from_str(expected_text14)
    assert score == expected_score


from musiclang.library import *

def test_extension_notes_basics():

    chord = (I % I.M)
    assert chord.to_pitch(b0) == 0
    chord = (I % I.M)['6']
    assert chord.to_pitch(b0) == 4
    assert chord.to_pitch(b1) == 7
    assert chord.to_pitch(b2) == 12
    assert chord.to_pitch(b3) == 16
    assert chord.to_pitch(b0.o(1)) == 16
    assert chord.to_pitch(b0.o(-1)) == -8



def test_extension_notes_complicated_chord():

    chord = (I % I.M)['2[add2]']
    assert chord.to_pitch(b0) == -1
    assert chord.to_pitch(b1) == 0
    assert chord.to_pitch(b2) == 2
    assert chord.to_pitch(b3) == 4


def test_extension_notes_to_scale_pitches_with_octave():
    chord = (I % I.M.o(1)).o(1)
    pitch = chord.to_pitch(b0)

    assert pitch == 24





from musiclang.library import *
from musiclang.transform.library import *
from musiclang.transform.composing.pattern import Nocturne, StandardPlucked, EpicMusicTernary
from musiclang.transform.composing import PartComposer
from musiclang.predict import compose_melody_and_harmony


def get_nocturne_orchestration(melody, harmony, tonality, bar_duration=4):
    composer = PartComposer(**{
        'orchestral_layers': [{
            "instruments": ['piano', 'piano', 'piano', 'piano'],
            "voicing": [b0.o(-1), b2.o(-1), b0, b1],
            "orchestra": Nocturne().to_json(),
            "fixed_bass": True,
            "voice_leading": True
        },

        ],
        'melodic_layers': [{
            "instrument": "piano",
            "melody": melody
        }],
        'global_layer': {
            "tonality": tonality,
            "harmony": harmony,
            "acc_amp": 'p',
            "patternator": {"restart_each_chord": False}

        }
    })
    return composer.to_json()


from musiclang.transform import TransformGraph



def test_multioutput():
    g = TransformGraph(inputs=['A', 'B', 'C'], outputs=('D', 'E'))
    g[('D', 'E')] = (('A', 'B'), (lambda a, b: (a + b, b)))
    res = g(1, 2, 3)
    assert res == (3, 2)



def test_singleoutput():
    g = TransformGraph(inputs=['A', 'B', 'C'], outputs='D')
    g[('D', 'E')] = (('A', 'B'), (lambda a, b: (a + b, b)))
    res = g(1, 2, 3)
    assert res == 3


import re
import itertools


CHORD_REGEX = re.compile(r'\([^)]+%[^)]+\)')


class ChordTokenizer:
    """Chord tokenizer transforms a score in text format to a list of tokens"""

    def tokenize(self, text):
        """Tokenize a text (or score) to extract chord tokens

        Parameters
        ----------
        text :
            Score or str, if Score first convert it to its string representation

        Returns
        -------
        tokens: List[str]
                List of tokens

        """
        chords = re.findall(CHORD_REGEX, str(text))
        # Deduplicate the chords, because we don't care about rythm here
        deduplicated_chords = [k for k, _g in itertools.groupby(chords)]
        return deduplicated_chords


    def tokenize_all(self, texts):
        """Tokenize all texts

        Parameters
        ----------
        texts :
            return: List of List of tokens

        Returns
        -------
        tokens_list: List[List[str]]
            A list of token per text

        """
        data = []
        for text in texts:
            data.append(self.tokenize(text))

        return data

    def tokenize_file(self, file):
        """Tokenize one file

        Parameters
        ----------
        file : str
            Filepath where to get the tokens

        Returns
        -------
        tokens: List[str]
                List of tokens

        """
        with open(file, 'r') as f:
            return self.tokenize(f.read())

    def tokenize_directory(self, directory):
        """
        Tokenize a directory, call the tokenize_file method for each text in the directory.

        Parameters
        ----------
        directory : str
            Directory to tokenize
            

        Returns
        -------

        data: List[List[tokens]]
             A list of token per text

        """
        import os
        files = [os.path.join(dirpath, file)
                 for (dirpath, dirnames, filenames) in os.walk(directory)
         for file in filenames]

        data = []
        for file in files:
            data.append(self.tokenize_file(file))

        return data


class ChordDetokenizer:
    """Convert tokens from chord tokenizer to chords"""

    def detokenize(self, tokens):
        """
        Convert a tokens list to a musiclang score.

        Parameters
        ----------
        tokens :
            

        Returns
        -------
        score: musiclang.Score
               Score detokenized (only chords)

        """
        from musiclang.write.library import I, II, III, IV, V, VI, VII
        return sum(eval('+'.join(tokens)), None)



import glob

import os
import multiprocessing as mp
import random

from musiclang import Score

import concurrent.futures
import time
import signal

def signal_handler(signum, frame):
    raise Exception("Timed out!")

signal.signal(signal.SIGALRM, signal_handler)


def extract_one(self, filename):
    try:
        score = self.extract_musiclang(filename)
        if self.remove_drums:
            score = Score([self.remove_drums_in_score(chord) for chord in score.chords])
            score = Score([chord for chord in score.chords if chord.duration > 0])
        if score is not None:
            text_filename = ''.join(os.path.basename(filename).split('.')[:-1]) + '.txt'
            pickle_filename = ''.join(os.path.basename(filename).split('.')[:-1]) + '.pkl'
            text_output_filename = os.path.join(self.text_output_directory, text_filename)
            pickle_output_filename = os.path.join(self.pickle_output_directory, pickle_filename)
            score.to_text_file(text_output_filename)
            score.to_pickle(pickle_output_filename)
            print('SUCCESS !')
    except Exception as e:
        print('Error while processing file %s' % filename)
        print(e)
        # Add the file to the error file
        with open(self.error_file, 'a') as f:
            f.write(os.path.basename(filename) + '\n')

        return None
def extract_one_with_time_limit(self, filename):
    import time
    print('Processing file %s' % filename)
    start = time.time()
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(self.time_budget_per_file)
    try:
        extract_one(self, filename)
    except Exception as msg:
        print("Timed out!", msg)
        with open(self.error_file, 'a') as f:
            f.write(os.path.basename(filename) + '\n')
    # Try to extract the file (self.extract_one) within a given time budget (self.time_budget_per_file)
    # If it fails, add the file to the error file
    print('TOOK', time.time() - start)

class DatasetExtractor:

    """
    Main class to extract musiclang notation from a glob pattern of midi files
    """

    def __init__(self, input_pattern, output_directory,
                 remove_drums=False, fast_chord_inference=True,
                 time_budget_per_file=120
                 ):

        self.remove_drums = remove_drums
        self.time_budget_per_file = time_budget_per_file
        self.fast_chord_inference = fast_chord_inference

        # List all the files to process
        self.files = glob.glob(input_pattern)
        random.shuffle(self.files)

        # Create output directory if it does not exist
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)
        # Create text output directory
        self.text_output_directory = os.path.join(output_directory, 'text')
        if not os.path.exists(self.text_output_directory):
            os.makedirs(self.text_output_directory)

        # Create pickle output directory
        self.pickle_output_directory = os.path.join(output_directory, 'pickle')
        if not os.path.exists(self.pickle_output_directory):
            os.makedirs(self.pickle_output_directory)

        # Create error file
        self.error_file = os.path.join(output_directory, 'errors.txt')

        # Filter files that have already been processed
        print('Original length :', len(self.files))
        self.filter_files()


    def filter_files(self):
        # Read error files and remove them from the list of files to process
        if os.path.exists(self.error_file):
            with open(self.error_file, 'r') as f:
                errors = f.read().split('\n')
            self.files = [file for file in self.files if not os.path.basename(file) in errors]
        print('Filtered error length :', len(self.files))
        self.files = [file for file in self.files
                      if not
                      os.path.exists(os.path.join(self.text_output_directory,
                                                  ''.join(os.path.basename(file).split('.')[:-1]) + '.txt'))]
        print('Filtered length :', len(self.files))

    def remove_drums_in_score(self, chord):
        return chord(**{ins: val for ins, val in chord.items() if not ins.startswith('drums')})

    def extract_musiclang(self, filename):
        """
        Extract musiclang notation from a midi file
        """
        return Score.from_midi(filename, fast_chord_inference=self.fast_chord_inference)



    def extract_all_files(self, n_jobs=1):
        """
        Multiprocess the extract one method to parallelize the extraction, it call extract_one with a mapping
        """
        import functools
        if n_jobs > 1:
            print('Using multiprocessing')
            pool = mp.Pool(n_jobs, maxtasksperchild=1)
            pool.map(functools.partial(extract_one_with_time_limit, self), self.files)
            pool.close()
            pool.join()
        else:
            print('Not using multiprocessing')
            import time

            for filename in self.files:
                print('Processing file %s' % filename)

                start = time.time()
                signal.signal(signal.SIGALRM, signal_handler)
                signal.alarm(self.time_budget_per_file)
                try:
                    extract_one(self, filename)
                except Exception as msg:
                    print("Timed out!", msg)
                    with open(self.error_file, 'a') as f:
                        f.write(os.path.basename(filename) + '\n')
                # Try to extract the file (self.extract_one) within a given time budget (self.time_budget_per_file)
                # If it fails, add the file to the error file
                print('TOOK', time.time() - start)




class FakeModel:

    def predict(self, start="Time Signature: 4/4\nm0", bar_number_start=0, seed=None, **config):
        text = """Time Signature: 4/4
                m0 C: I b3 V7
                m1 IV b3 V7
                m2 I
        """

        text = '\n'.join([c.lstrip() for c in text.split('\n')])
        return text


model = FakeModel()

from musiclang import Score
from musiclang.predict import AutoComposer

expected_annotation = """Time Signature: 4/4
mx C: I b3 V7
mx IV b3 V7
mx C: I b3 V7
mx IV b3 V7
mx I
mx C: I b3 V7"""

expected_text = """
(I % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V['7'] % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(IV % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V['7'] % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(I % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V['7'] % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(IV % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V['7'] % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(I % I.M)(
	piano__0=b0.w,
	piano__1=b1.w,
	piano__2=b2.w,
	piano__3=b3.w)+
(I % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)+
(V['7'] % I.M)(
	piano__0=b0.h,
	piano__1=b1.h,
	piano__2=b2.h,
	piano__3=b3.h)
"""
def test_auto_composer():
    orch4 = """
    """

    orchestras = [orch4]
    parts = [2, 4]
    tonalities = ['d', 'a']
    form = [0, 1]
    form_orchestra = [0, 0]
    signature = (4, 4)
    ###################

    composer = AutoComposer(model, parts, tonalities, form, form_orchestra, orchestras, signature=signature, continuation=False)
    score, annotation = composer.compose()
    expected_score = Score.from_str(expected_text)
    assert annotation == expected_annotation
    assert score == expected_score





_TOKENIZER = None
_MODEL = None
_GPU = False
_DEVICE = None
def _load_models():
    global _TOKENIZER, _MODEL, _GPU, _DEVICE
    from transformers import GPT2LMHeadModel, GPT2TokenizerFast
    import torch
    if _TOKENIZER is not None:
        return _TOKENIZER, _MODEL
    hub_model_path = "floriangardin/musiclang"
    TOKENIZER = GPT2TokenizerFast.from_pretrained(hub_model_path, padding_side='left')
    MODEL = GPT2LMHeadModel.from_pretrained(hub_model_path)
    _GPU = torch.cuda.is_available()
    _DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    if _GPU:
        MODEL = MODEL.to('cuda')
    _TOKENIZER = TOKENIZER
    _MODEL = MODEL

    return TOKENIZER, MODEL



def preprocess(prompt, tokenizer):
    """
    Preprocess the prompt stripping spaces, line return and tabulations
    Parameters
    ----------
    prompt

    Returns
    -------

    """
    prompt = prompt.replace(' ', '').replace('\t', '').replace('\r', '').replace('\n', '')
    prompt = tokenizer.eos_token + prompt
    return prompt




def generate_one_chord(input_text, model, tokenizer,
                                 max_length=100, temperature=0.7,
                                 top_k=20,
                                 target_sequence=')+('):
    # Tokenize the input text
    import torch
    block_size = 1024
    input_ids = tokenizer.encode(input_text, return_tensors='pt').to(_DEVICE)
    target_ids = tokenizer.encode(target_sequence)

    target_length = len(target_ids)
    start_length = len(input_ids[0])

    # split between block size and predicted sequence
    # Divide start_length between block_size - max_length
    end_ids = input_ids[:, -(block_size-max_length):]
    start_ids = input_ids[:, :-(block_size-max_length)]
    assert len(end_ids[0]) + len(start_ids[0]) == start_length
    idx = 0
    while True:
        # Generate text with the model
        idx += 1
        end_ids = model.generate(
            end_ids,
            max_new_tokens=max_length,
            num_return_sequences=1,
            do_sample=True,
            temperature=temperature,
            top_k=top_k,
            bad_words_ids=[[tokenizer.eos_token_id]],
            pad_token_id=tokenizer.eos_token_id,
        )

        output = torch.cat((start_ids, end_ids), 1)
        end_ids = output[:, -(block_size - max_length):]
        start_ids = output[:, :-(block_size - max_length)]
        # Decode the generated text
        output_text = tokenizer.decode(output[0], skip_special_tokens=False)
        output_text = output_text.replace(tokenizer.eos_token, '+')
        if output_text.startswith('+'):
            output_text = output_text[1:]
        if target_sequence in output_text[start_length+target_length:]:
            output_text = target_sequence.join(output_text.split(target_sequence)[:-1]) + ')'
            return output_text


import re


def deduplicate_chord_instruments(text):
    # Split the text into parts
    parts = re.split('([a-zA-Z0-9_]+=)', text)

    # Create a new list to hold the modified parts
    modified_parts = []

    # Keep track of the keys we've seen
    seen_keys = {}

    # Process each part
    for part in parts:
        if part.endswith('='):
            # This part is a key, check if we've seen it before
            key = part[:-1]  # remove the '='
            ins, idx = key.split('__')
            idx = int(idx)
            modified_key = None
            if (ins, idx) in seen_keys:
                # We've seen this key before, append a unique identifier
                while (ins, idx) in seen_keys:
                    ind, idx = ins, idx + 1
                    modified_key = ins + '__' + str(idx)

                seen_keys[(ins, idx)] = True
            else:
                # This is a new key, add it to the dictionary
                modified_key = key
                seen_keys[(ins, idx)] = True

            # Add the modified key to the list
            modified_parts.append(modified_key + '=')
        else:
            # This part is not a key, just add it to the list
            modified_parts.append(part)

    # Join the parts back together and return
    return ''.join(modified_parts)


def generate_n_chords(text, model, tokenizer, n, **kwargs):
    """
    Generate n chords from a text prompt
    Parameters
    ----------
    text: str
    model: GPT2LMHeadModel

    tokenizer: GPT2TokenizerFast
    n: int
        The number of chords to generate


    Returns
    -------
    text: str
        The generated text
    """
    for i in range(n):
        text = generate_one_chord(text, model, tokenizer, **kwargs)
        chords = text.split(')+(')
        chords[-1] = deduplicate_chord_instruments(chords[-1])
        text = ')+('.join(chords)
    return text

def predict_score_from_huggingface(prompt, n_chords=1, temperature=0.75, top_k=20):
    """
    Given a text prompt, generate a score using the huggingface model

    Parameters
    ----------
    prompt: str
        The text prompt
    n_chords: int
        The number of chords to generate
    temperature: float
        The temperature of the generation, the higher the more random, usually don't go over 1.0 to avoid syntax errors
    top_k: int
        The number of top k tokens to consider when generating the next token


    Returns
    -------
    result: str
        The generated text score
    """


    TOKENIZER, MODEL = _load_models()

    prompt = preprocess(prompt, TOKENIZER)


    return generate_n_chords(prompt, MODEL, TOKENIZER, n_chords, temperature=temperature, top_k=top_k)



from fractions import Fraction as frac
import musiclang.library as L

def realize_tags(note, last_note=None, next_note=None):
    """
    Given the tags of a given note returns a new melody with the realized tags
    Parameters
    ----------
    note
    last_note
    next_note

    Returns
    -------

    """
    import musiclang.library as L
    new_note = note.copy()
    duration = note.duration
    if 'accent' in note.tags:
        new_note = accent(new_note, last_note, next_note)

    if 'mordant' in note.tags:
        new_note = mordant(new_note, last_note, next_note)

    if 'inv_mordant' in note.tags:
        new_note = inv_mordant(new_note, last_note, next_note)

    if 'chroma_mordant' in note.tags:
        new_note = chroma_mordant(new_note, last_note, next_note)


    if 'inv_chroma_mordant' in note.tags:
         new_note = inv_chroma_mordant(new_note, last_note, next_note)

    if 'grupetto' in note.tags:
        new_note = grupetto(new_note, last_note, next_note)

    if 'inv_grupetto' in note.tags:
        new_note = inv_grupetto(new_note, last_note, next_note)

    if 'chroma_grupetto' in note.tags:
        new_note = chroma_grupetto(new_note, last_note, next_note)

    if 'inv_chroma_grupetto' in note.tags:
        new_note = inv_chroma_grupetto(new_note, last_note, next_note)

    if 'roll' in note.tags:
        new_note = roll(new_note, last_note, next_note)

    if 'roll_fast' in note.tags:
        new_note = roll_fast(new_note, last_note, next_note)

    if 'suspension_prev' in note.tags:
        new_note = suspension(new_note, last_note, next_note)

    if 'suspension_prev_repeat' in note.tags:
        new_note = suspension_prev_repeat(new_note, last_note, next_note)

    if 'retarded' in note.tags:
        new_note = retarded(new_note, last_note, next_note)


    if 'interpolate' in note.tags:
        new_note = interpolate(new_note, last_note, next_note)

    assert new_note.duration == note.duration, f"{new_note} {new_note.duration} {note.duration} {note.tags}"
    new_note = new_note.clear_note_tags()
    return new_note



def accent(new_note, last_note, next_note):
    new_note.amp = min(120, new_note.amp + 10)
    return new_note

def mordant(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.set_duration(mordant_duration) + L.su1.set_duration(
            mordant_duration) + new_note.set_duration(new_note.duration - 2 * mordant_duration)
    return new_note

def inv_mordant(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.set_duration(mordant_duration) + L.sd1.set_duration(
            mordant_duration) + new_note.set_duration(new_note.duration - 2 * mordant_duration)
    return new_note


def chroma_mordant(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.set_duration(mordant_duration) + L.hu1.set_duration(
            mordant_duration) + new_note.set_duration(new_note.duration - 2 * mordant_duration)
    return new_note

def inv_chroma_mordant(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.set_duration(mordant_duration) + L.hd1.set_duration(
            mordant_duration) + new_note.set_duration(new_note.duration - 2 * mordant_duration)
    return new_note


def grupetto(new_note, last_note, next_note):
    duration = new_note.duration
    if duration >= frac(1):
        mordant_duration = frac(1, 2)
        new_note = new_note.n + L.su1.set_duration(mordant_duration) + new_note.set_duration(mordant_duration) \
                   + L.sd1.set_duration(mordant_duration) + L.su1.set_duration(duration - 3 * mordant_duration)
    elif duration >= frac(1, 2):
        mordant_duration = frac(2, 3) * frac(1, 4)
        new_note = new_note.n + L.su1.set_duration(mordant_duration) + new_note.set_duration(mordant_duration) \
                   + L.sd1.set_duration(mordant_duration) + L.su1.set_duration(duration - 3 * mordant_duration)
    return new_note

def inv_grupetto(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(2, 3) * frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.n + L.sd1.set_duration(mordant_duration) + new_note.set_duration(mordant_duration) \
                   + L.su1.set_duration(mordant_duration) + L.sd1.set_duration(duration - 3 * mordant_duration)

    return new_note

def chroma_grupetto(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(2, 3) * frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.n + L.hu1.set_duration(mordant_duration) + new_note.set_duration(mordant_duration) \
                   + L.hd1.set_duration(mordant_duration) + L.hu1.set_duration(duration - 3 * mordant_duration)
    return new_note


def inv_chroma_grupetto(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(2, 3) * frac(1, 4)
    if duration >= frac(1, 2):
        new_note = new_note.n + L.hd1.set_duration(mordant_duration) + new_note.set_duration(mordant_duration) \
                   + L.hu1.set_duration(mordant_duration) + L.hd1.set_duration(duration - 3 * mordant_duration)

    return new_note


def roll(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 4)
    nb_rolls = int(duration / mordant_duration)
    melody = None
    for i in range(nb_rolls):
        if (i % 2) == 0:
            melody += new_note.set_duration(mordant_duration)
        else:
            melody += L.su1.set_duration(mordant_duration)

    if nb_rolls != (duration / mordant_duration):
        melody += L.l(duration - nb_rolls * mordant_duration)
    new_note = melody

    return new_note

def roll_fast(new_note, last_note, next_note):
    duration = new_note.duration
    mordant_duration = frac(1, 6)
    nb_rolls = int(duration / mordant_duration)
    melody = None
    for i in range(nb_rolls):
        if (i % 2) == 0:
            melody += new_note.set_duration(mordant_duration)
        else:
            melody += L.su1.set_duration(mordant_duration)

    if nb_rolls != (duration / mordant_duration):
        melody += L.l(duration - nb_rolls * mordant_duration)
    new_note = melody
    return new_note


def suspension(new_note, last_note, next_note):
    duration = new_note.duration
    if last_note:
        new_note = L.l.set_duration(duration / 2) + new_note.set_duration(duration / 2)

    return new_note


def suspension_prev_repeat(new_note, last_note, next_note):
    duration = new_note.duration
    if last_note:
        new_note = last_note.set_duration(duration / 2) + new_note.set_duration(duration / 2)

    return new_note

def retarded(new_note, last_note, next_note):
    duration = new_note.duration
    retarded_duration = frac(1, 12)
    new_note = L.l.set_duration(retarded_duration) + new_note.set_duration(duration - retarded_duration)
    return new_note


def interpolate(new_note, last_note, next_note):
    """
    Interpolate note of the scale between current note and next_note
    Parameters
    ----------
    new_note: Current note to modify
    last_note: Previous note in the melody
    next_note: Next note in the melody

    Returns
    -------

    """
    from musiclang import Note
    if next_note is None or not next_note.is_note or not new_note.is_note:
        return new_note
    
    first_val = new_note.val if new_note.type == 's' else int(7 * new_note.val/12)
    first_val += 7 * new_note.octave
    second_val = next_note.val if next_note.type == 's' else int(7 * next_note.val / 12)
    second_val += 7 * next_note.octave
    delta_scale = second_val - first_val
    if delta_scale == 0:
        return new_note
    duration = frac(new_note.duration, int(abs(delta_scale)))

    up = delta_scale > 0
    temp_note = L.su1.set_duration(duration) if up else L.sd1.set_duration(duration)
    melody = None
    for i in range(first_val, second_val, 1 if up else -1):
        if i == first_val:
            melody += new_note.set_duration(duration)
        else:

            melody += temp_note
    melody =  melody.set_amp(new_note.amp)
    return melody



import joblib
import os
import pandas as pd
import numpy as np
from musiclang.write.elements import ChordElement

class Arranger:

    def __init__(self, directory, nb_candidates=None):
        self.directory = directory
        filepath_states = os.path.join(self.directory, 'states.pickle')
        filepath_priors = os.path.join(self.directory, 'priors.pickle')
        filepath_transitions = os.path.join(self.directory, 'transitions.pickle')
        filepath_emissions = os.path.join(self.directory, 'emissions.pickle')
        print('Loading infos ...')
        self.states = joblib.load(filepath_states)
        self.priors = self._to_series(joblib.load(filepath_priors))
        self.transitions = self._to_dataframe(joblib.load(filepath_transitions))
        self.emissions = self._to_dataframe(joblib.load(filepath_emissions))
        print('Finished loading')
        if nb_candidates is not None:
            # Find the most probable candidates
            keep_index = {mode: self.priors[mode].sort_values(ascending=False).index[:nb_candidates]
                          for mode in self.priors.keys()}
            self.transitions = {mode: self.transitions[mode].loc[keep_index[mode], keep_index[mode]]
                                for mode in self.transitions.keys()}
            self.emissions = {mode: self.emissions[mode].loc[keep_index[mode]]
                                for mode in self.transitions.keys()}
            self.priors = {mode: self.priors[mode].loc[keep_index[mode]]
                                for mode in self.priors.keys()}

            self.states = {mode: [s for s in self.states[mode] if s in keep_index[mode]] for mode in self.states.keys()}

    @staticmethod
    def _to_dataframe(x):
        return {key: pd.DataFrame(val) for key, val in x.items()}

    @staticmethod
    def _to_series(x):
        return {key: pd.Series(val) for key, val in x.items()}

    def arrange(self, melody, tonality, candidates=None, instrument='piano',
                temperature=0.0, eps=1e-5, zero_diag=True, weak_beat_importance=0.5,
                return_score=False
                ):

        # melody= [[(s0, 1), (s1, 0)], [], []]
        tone, mode = ChordElement.get_key_mode(tonality)
        # Get pitches from melody
        base_chord = ChordElement.get_tonic_chord(tonality)
        pitches = []
        curr_pitch = None
        arrange_with_next = False
        for chord_change in melody:
            temp_pitches = []
            for note, is_in_chord in chord_change:
                if is_in_chord and not note.is_silence:
                    curr_pitch = (base_chord.to_pitch(note, last_pitch=curr_pitch) - tone) % 12
                    temp_pitches.append((curr_pitch, 1.0))
                elif not is_in_chord and not note.is_silence:
                    curr_pitch = (base_chord.to_pitch(note, last_pitch=curr_pitch) - tone) % 12
                    temp_pitches.append((curr_pitch, weak_beat_importance))

            if arrange_with_next and len(temp_pitches) > 0:
                pitches = [list(temp_pitches)] * len(pitches)
                arrange_with_next = False

            if len(temp_pitches) == 0 and len(pitches) > 0:
                pitches.append(pitches[-1])

            elif len(temp_pitches) == 0 or arrange_with_next:
                pitches.append([])
                arrange_with_next = True

            if len(temp_pitches) > 0:
                pitches.append(temp_pitches)

        if candidates is None:
            candidates = [None] * len(pitches)

        chords, proba = self.get_chord_progression(pitches, candidates, tonality,
                                              self.states,
                                              self.priors,
                                              self.transitions,
                                              self.emissions,
                                              temperature=temperature,
                                              eps=eps,
                                              zero_diag=zero_diag)

        if not return_score:
            return chords
        # Re-arrange the melody inside the chords
        score = None
        for literal_chord, chord_change in zip(chords, melody):
            chord_melody = None
            chord = ChordElement(literal_chord).parse_with_tonality(tonality)
            for note, is_in_chord in chord_change:
                if note.is_note:
                    curr_pitch = base_chord.to_pitch(note, last_pitch=curr_pitch)
                    new_note = chord.parse(curr_pitch).set_duration(note.duration)
                else:
                    new_note = note.copy()
                chord_melody += new_note

            score += chord(**{instrument: chord_melody})


        return score



    def viterbi(self, obs, candidates, states, start_p, trans_p, emit_p, eps=1e-5):

        def get_emit(st, t):
            if isinstance(obs[t], int):
                # If observable is a single pitch
                emit = emit_p[st][obs[t]]
            else:
                # If multi pitches
                emit = 1.0
                for ob, strength in obs[t]:
                    emit *= emit_p[st][ob] * strength
            if candidates[t] is not None:
                if isinstance(candidates[t], str):
                    emit = 1.0 * (st == candidates[t]) + eps
                else:
                    emit = 1.0 * (st in candidates[t]) / len(candidates[t]) + eps

            return emit

        V = [{}]
        for st in states:
            V[0][st] = {"prob": start_p[st] * get_emit(st, 0), "prev": None}
        # Run Viterbi when t > 0
        for t in range(1, len(obs)):
            V.append({})
            for st in states:
                max_tr_prob = V[t - 1][states[0]]["prob"] * trans_p[states[0]][st] * get_emit(st, t)
                prev_st_selected = states[0]
                for prev_st in states[1:]:
                    tr_prob = V[t - 1][prev_st]["prob"] * trans_p[prev_st][st] * get_emit(st, t)
                    if tr_prob > max_tr_prob:
                        max_tr_prob = tr_prob
                        prev_st_selected = prev_st

                max_prob = max_tr_prob
                V[t][st] = {"prob": max_prob, "prev": prev_st_selected}

        opt = []
        max_prob = - np.inf
        best_st = None
        # Get most probable state and its backtrack
        for st, data in V[-1].items():
            if data["prob"] > max_prob:
                max_prob = data["prob"]
                best_st = st
        opt.append(best_st)
        previous = best_st

        # Follow the backtrack till the first observation
        for t in range(len(V) - 2, -1, -1):
            opt.insert(0, V[t + 1][previous]["prev"])
            previous = V[t + 1][previous]["prev"]
        return opt, max_prob

    def dptable(self, V):
        # Print a table of steps from dictionary
        yield " " * 5 + "     ".join(("%3d" % i) for i in range(len(V)))
        for state in V[0]:
            yield "%.7s: " % state + " ".join("%.7s" % ("%lf" % v[state]["prob"]) for v in V)

    def get_chord_progression(self, pitches, candidates, tonality, states, priors, transitions,
                              emissions, temperature=0.0, eps=1e-5, zero_diag=True):
        # Get pitch relative to tonality
        tone, mode = ChordElement.get_key_mode(tonality)
        #obs = [((p[0] - tone) % 12, p[1]) if isinstance(p[0], tuple) else [((pi[0] - tone), pi[1]) % 12 for pi in p] for p in pitches]
        obs = pitches
        # If temperature add random noise to each matrix
        if temperature > 0:
            priors = {m: prior + (temperature * np.random.randn(*prior.shape)) for m, prior in priors.items()}
            transitions = {m: transition + (temperature * np.random.randn(*transition.shape)) for m, transition in transitions.items()}
            emissions = {m: emission + (temperature * np.random.randn(*emission.shape)) for m, emission in emissions.items()}

        if zero_diag:
            for mode_ in transitions.keys():
                transitions[mode_] = transitions[mode_] * (1 - np.eye(*transitions[mode_].shape))
        chords, prob = self.viterbi(obs,
                               candidates,
                               states[mode],
                               (priors[mode]).to_dict(),
                               (transitions[mode].T + eps).to_dict(),
                               (emissions[mode].T + eps).to_dict(),
                               eps=eps)
        return chords, prob




from musiclang import Metric
from musiclang.library import *
import pytest

def test_basic_metric_instantiation_ok():

    metric = Metric([1, 0, 0, 0], signature=(4, 4), tatum=1)
    assert metric.tatum == 1
    assert metric.signature == (4, 4)
    assert metric.nb_bars == 1
    assert metric.array == [1, 0, 0, 0]



def test_error_if_array_does_not_match_duration_():

    with pytest.raises(ValueError):
        metric = Metric([1, 0, 0], signature=(4, 4), tatum=1)


def test_strong_beat_four_four_():

    metric = Metric([1, 0, 0, 0], signature=(4, 4), tatum=1)
    assert metric.beat_type(0) == Metric.STRONG
    assert metric.beat_type(1) == Metric.WEAK
    assert metric.beat_type(0.5) == Metric.SYNCOPATION


def test_apply_melody_with_expand():
    melody = s0
    metric = Metric([1, 0, 1, 0], signature=(4, 4), tatum=1)
    res = metric.apply_to_melody(melody, expand=True)
    assert res == s0.h + s0.h


def test_apply_melody_with_expand_two_notes():
    melody = s0 + s1.e
    metric = Metric([1, 0, 1, 0], signature=(4, 4), tatum=1)
    res = metric.apply_to_melody(melody, expand=True)
    assert res == s0.h + s1.h


def test_apply_melody_without_expand():

    melody = s0
    metric = Metric([1, 0, 1, 0], signature=(4, 4), tatum=1)
    res = metric.apply_to_melody(melody, expand=False)
    assert res == s0.h + r.h


def test_canon():
    metric = Metric([1, 0, 1, 0], signature=(4, 4), tatum=1)
    canon = metric.canon(2)
    assert canon.array == [0, 0, 1, 0]

def test_circular_shift():
    metric = Metric([1, 0, 1, 0], signature=(4, 4), tatum=1)
    canon = metric.circular_shift(2)
    assert canon.array == [1, 0, 1, 0]


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from ..constants import CONTINUATION, SILENCE
from ..note_pitch import NotePitch
from ..pitches.pitches_utils import note_to_pitch_result


def get_track_list(score):
    """

    Parameters
    ----------
    score :
        

    Returns
    -------

    """
    items = sum([chord.parts for chord in score.chords], [])
    tracks = list(dict.fromkeys(items))

    return tracks


def get_or_default_last_pitch(chord, track_idx, time, last_pitch):
    """

    Parameters
    ----------
    chord :
        
    track_idx :
        
    time :
        
    last_pitch :
        

    Returns
    -------

    """
    if last_pitch is not None:
        return last_pitch

    return None


def note_to_pitch(note, chord, track_idx, time, last_pitch):
    """

    Parameters
    ----------
    note :
        
    chord :
        
    track_idx :
        
    time :
        
    last_pitch :
        

    Returns
    -------

    """

    # Get real scale pitches
    if last_pitch is not None:
        last_pitch_pitch = last_pitch[0]
    else:
        last_pitch_pitch = 0

    pitch_result = note_to_pitch_result(note, chord, last_pitch=last_pitch_pitch)

    if pitch_result is None:
        pitch_result = 0

    pitch = NotePitch(pitch_result, offset=time, duration=note.duration, track=track_idx,
                      velocity=note.amp,
                      silence=1 * (note.is_silence or (note.is_continuation and last_pitch is None)),
                      continuation=1 * (note.is_continuation and last_pitch is not None),
                      tempo= note.tempo,
                      pedal= note.pedal
                      )

    if pitch.is_note():
        last_pitch = pitch.to_midi_note()

    return pitch.to_midi_note(), last_pitch


def melody_to_pitches(part, chord, track_idx, time, last_pitch):
    """

    Parameters
    ----------
    part :
        
    chord :
        
    track_idx :
        
    time :
        
    last_pitch :
        

    Returns
    -------

    """
    chord_result = []
    for note in part.notes:
        to_append, last_pitch = note_to_pitch(note, chord, track_idx, time, last_pitch)
        time += note.duration
        chord_result.append(to_append)

    return chord_result, last_pitch


def create_melody_for_track(score, track_idx, track):
    """

    Parameters
    ----------
    score :
        
    track_idx :
        
    track :
        

    Returns
    -------

    """
    result = []
    last_pitch = None
    time = 0
    from musiclang import Melody, Silence
    for idx_chord, chord in enumerate(score.chords):
        if track in chord.score.keys():
            last_pitch = get_or_default_last_pitch(chord, track_idx, time, last_pitch)
            part = chord.score[track]
            last_note = score.chords[idx_chord - 1].score.get(track, Melody([Silence(1)])).notes[-1] if idx_chord -1 >= 0 else None
            next_note = score.chords[idx_chord + 1].score.get(track, Melody([Silence(1)])).notes[0] if idx_chord + 1 < len(score.chords) else None
            part_enriched = part.realize_tags(last_note=last_note, final_note=next_note)
            chord_result, last_pitch = melody_to_pitches(part_enriched, chord, track_idx, time, last_pitch)
            result += chord_result
        else:
            last_pitch = None
        time += chord.duration
    return result


def to_midi(notes, output_file=None, **kwargs):
    """

    Parameters
    ----------
    notes :
        
    output_file :
         (Default value = None)
    **kwargs :
        

    Returns
    -------

    """
    from .midi_utils import matrix_to_mid
    res = matrix_to_mid(notes, output_file=output_file, **kwargs)
    return res


def tracks_to_instruments(tracks):
    """

    Parameters
    ----------
    tracks :
        

    Returns
    -------

    """
    from .constants import INSTRUMENTS_DICT
    names = [t.split('__')[0] for t in tracks]
    instruments_idx = {i: INSTRUMENTS_DICT.get(name, 0) for i, name in enumerate(names)}
    return instruments_idx, names




def get_pitches_instruments(score):
    tracks = get_track_list(score)

    # For each track create melody
    melodies = []
    for track_idx, track in enumerate(tracks):
        melody = create_melody_for_track(score, track_idx, track)
        melodies.append([n[0] for n in melody if not n[CONTINUATION] and not n[SILENCE]])

    return melodies
def get_pitches(score):
    notes = get_notes(score)
    return [n[0] for n in notes if not n[CONTINUATION] and not n[SILENCE]]

def get_notes(score):
    """
    Get all notes from a score

    Parameters
    ----------
    score: Score
        The score to get notes from

    Returns
    -------
    notes: list
        A list of notes

    """
    # Get all tracks
    tracks = get_track_list(score)

    # For each track create melody
    notes = []
    for track_idx, track in enumerate(tracks):
        melody = create_melody_for_track(score, track_idx, track)
        notes += melody

    return notes



def score_to_midi(score, filepath, **kwargs):
    """Transform a score to a midi file

    Parameters
    ----------
    score :
        param filepath:
    kwargs :
        return:
    filepath :
        
    **kwargs :
        

    Returns
    -------

    """

    # Get all tracks
    tracks = get_track_list(score)

    # For each track create melody
    notes = get_notes(score)
    instruments, instrument_names = tracks_to_instruments(tracks)

    res = to_midi(notes, output_file=filepath, instruments=instruments, instrument_names=instrument_names, **kwargs)
    return res


def score_to_events(score, **kwargs):
    """Transform a score to a event list

    Parameters
    ----------
    score :
        param filepath:
    kwargs :
        return:
    filepath :

    **kwargs :


    Returns
    -------

    """

    # Get all tracks
    tracks = get_track_list(score)

    # For each track create melody
    notes = get_notes(score)
    instruments, instrument_names = tracks_to_instruments(tracks)

    res = matrix_to_events(notes, instruments=instruments, instrument_names=instrument_names, **kwargs)
    return res

def matrix_to_events(matrix,
                     ticks_per_beat=480,
                     tempo=120,
                     instruments={},
                     time_signature=(4, 4),
                     instrument_names=None, **kwargs):
    matrix = list(sorted(matrix, key=lambda x: x[1]))
    events = {}  # [(time, instrument, duration, message)]

    for idx, el in enumerate(matrix):
        pitch, offset, duration, velocity, track, silence, continuation, tempo_change, pedal = el
        # Add note on, note off for each track
        if tempo_change is not None:
            tempo = tempo_change

        real_instrument_name = instrument_names[track]
        second_offset = offset * 60 / tempo
        second_duration = float(duration * 60 / tempo)

        if not continuation:
            event = {
                'event_name': 'note_on',
                'duration': float(second_duration),
                'pitch': int(pitch),
                'offset': float(second_offset),
                'velocity': int(velocity),
                'instrument': real_instrument_name,
                'silence': silence,
                'pedal': pedal
            }

            events[track] = events.get(track, []) + [event]
        if continuation:
            last_event = events.get(track, None)
            if last_event is not None:
                last_event[-1]['duration'] += duration
            else:
                event = {
                    'event_name': 'note_on',
                    'duration': float(second_duration),
                    'pitch': int(pitch),
                    'offset': float(second_offset),
                    'velocity': int(velocity),
                    'instrument': real_instrument_name,
                    'silence': True,
                    'pedal': pedal
                }
                events[track] = events.get(track, []) + [event]

    events = sum([ev for key, ev in events.items()], [])
    events = [e for e in events if not e['silence']]
    events = list(sorted(events, key=lambda x: x['offset']))

    return events


from musiclang.transform.merger import ScoreMerger


class ConcatScores(ScoreMerger):

    def action(self, *scores, **kwargs):
        return sum(scores, None)


class TakeFirst(ScoreMerger):

    def action(self, *scores, **kwargs):
        return scores[0]


class TakeLast(ScoreMerger):

    def action(self, *scores, **kwargs):
        return scores[-1]


class TakeIdx(ScoreMerger):

    def __init__(self, idx, **kwargs):
        self.idx = idx

    def action(self, *scores, **kwargs):
        return scores[self.idx]


class TakeSlice(ScoreMerger):

    def __init__(self, slice, **kwargs):
        self.slice = slice

    def action(self, *scores, **kwargs):
        return scores[self.slice]


class ConcatWithPattern(ScoreMerger):

    def __init__(self, pattern, **kwargs):
        self.pattern = pattern

    def action(self, *scores, **kwargs):
        return sum([scores[p] for p in self.pattern], None)


class RandomSelector(ScoreMerger):
    def __init__(self, rg=None):
        import numpy as np
        self.rg = rg if rg is None else np.random.RandomState()

    def __call__(self, *scores, **kwargs):
        return self.rg.choice(scores)


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import autopep8

def chord_serie_to_code(chord_serie, **kwargs):
    """Export a string with all the generated python code

    Parameters
    ----------
    chord_serie :
        
    **kwargs :
        

    Returns
    -------

    """
    result = []
    result += get_import_lines()
    result += line_return()
    result += get_initial_variable_declaration(**kwargs)
    result += line_return()

    chord_serie_result = []
    for chord_idx, chord in enumerate(chord_serie.chords):

        chord_result = []
        for part_name in chord.parts:
            part = chord.score[part_name]
            code, variable_name = get_part_variables(part, chord_idx, part_name)
            result += code
            chord_result.append((code, variable_name, part_name))

        chord_code, chord_variable_name = get_chord_variable(chord_idx, chord, chord_result)
        chord_serie_result.append(chord_variable_name)
        result += chord_code
        result += line_return()

    # Get chord serie
    result += line_return()

    result += get_score(chord_serie_result)
    result += line_return()
    result += get_end_result()


    result = "\n".join(result)

    result = autopep8.fix_code(result
                                )

    return result

def line_return():
    """ """
    return ['']

def get_score(chord_serie_result):
    """

    Parameters
    ----------
    chord_serie_result :
        

    Returns
    -------

    """
    score_value = " + ".join(chord_serie_result)
    score = [f"score = {score_value}"]
    return score


def get_chord_variable(chord_idx, chord, chord_result):
    """

    Parameters
    ----------
    chord_idx :
        
    chord :
        
    chord_result :
        

    Returns
    -------

    """
    chord_variable_name = f"chord_{chord_idx}"
    code_inside_chord = ""
    for code_part, part_variable_name, part_name in chord_result:
        code_inside_chord += f"{part_name}={part_variable_name},"

    code_inside_chord = code_inside_chord[:-1]

    code_chord_def = chord.to_code()
    code = f"{code_chord_def}({code_inside_chord})"
    result = [f"{chord_variable_name} = {code}"]
    return result, chord_variable_name



def get_part_variables(part, chord_idx, part_name):
    """

    Parameters
    ----------
    part :
        
    chord_idx :
        
    part_name :
        

    Returns
    -------

    """
    variable_name = part_name + '_' + str(chord_idx)

    code = part.to_code()

    result = [f"{variable_name} = {code}"]

    return result, variable_name


def get_end_result():
    """ """
    result = ["score.to_midi(FILENAME, tempo=TEMPO)"]
    return result

def get_initial_variable_declaration(tempo=120, filename="", **kwargs):
    """

    Parameters
    ----------
    tempo :
         (Default value = 120)
    filename :
         (Default value = "")
    **kwargs :
        

    Returns
    -------

    """
    result = [
        "### GLOBAL SCORE VARIABLES",
        "",
        "TEMPO = {tempo}".format(tempo=tempo),
        f'FILENAME = "{filename}" # REPLACE WITH YOUR DESTINATION',
        "",
        "###"
              ]
    return result

def get_import_lines():
    """ """
    result =[
            "from musiclang.write.library import *",
            "from fractions import Fraction as frac"
             ]

    return result





"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from musiclang import Melody


def put_on_same_chord(score):
    """Take the first chord as reference,
    Put everything into the first chord preserving the note value (It will change the piece to a static harmony)

    Parameters
    ----------
    score : Score
            Input score
        

    Returns
    -------


    """
    from musiclang import Silence
    first_chord = score[0]
    instruments = score.instruments
    parts = {ins: None for ins in instruments}

    for chord in score.chords:
        duration = chord.duration
        for instrument in instruments:
            parts[instrument] += chord.score.get(instrument, Silence(duration))

    return first_chord(**parts)


def project_on_score(score, score2, keep_score=False):
    """Project harmonically the score onto the score2
    Algorithm : For each chord of score2 : get chords that belongs to score1 and reproject on chord of score2

    Parameters
    ----------
    score : Score
            Score to project on the chords of score2
    score2 : Score
        Score that contains the harmony
    keep_score : Score (Default value = False)
        Keep the voices of score2 ? (Default value = False)

    Returns
    -------
    new_score: Score
              The score projected on score2 chord progression

    """

    start_time = 0
    new_score = None
    #
    instruments1 = [(ins.split('__')[0], int(ins.split('__')[1])) for ins in score.instruments]

    for idx, chord2 in enumerate(score2):
        # Get all segments of chord
        end_time = start_time + chord2.duration

        subscore1 = score.get_score_between(start_time, end_time)
        if subscore1 is None:
            break

        subscore = subscore1.put_on_same_chord()
        #assert subscore.duration == chord2.duration, "Wrong duration when projecting"


        if keep_score:
            chord_score = chord2.score
            chord_score.update(subscore.score)
        else:
            chord_score = subscore.score
        new_score += chord2(**chord_score)

        start_time = end_time

    return new_score


def get_chord_between(chord, start, end):
    """Get the chord with melodies that are between start and end

    Parameters
    ----------
    chord :
        chord on which to extract subchord melodies
    start :
        Start time (relative to chord)
    end :
        End time (relative to chord)

    Returns
    -------
    type
        chord, Chord, modified chord

    """
    new_parts = {ins: None for ins in chord.instruments}
    for part in chord.score.keys():
        time = 0
        voice = chord.score[part]
        new_voice = []
        to_break = False
        for note in voice:
            add_continuation = False
            note_duration = note.duration
            new_note = note.copy()
            if time >= end:
                break
            if (time < start) and (time + note_duration <= start):
                time += new_note.duration
                continue
            if time + note_duration >= end:
                new_note.duration = end - time
                to_break = True
            if time < start:
                new_note.duration = new_note.duration - (start - time)
                time += start - time
                add_continuation = True

            if add_continuation:
                from ..note import Continuation
                new_voice.append(Continuation(new_note.duration))
            else:
                new_voice.append(new_note)


            if new_note.duration < 0:
                raise Exception('Get a negative duration in get_chord_between')


            time += new_note.duration

            if to_break:
                break
        new_voice = Melody(new_voice)
        new_parts[part] = new_voice

    if len(new_parts.keys()) == 0:
        from ..note import Silence
        return chord(**{'piano__0': Silence(end - start)})
        # if new_voice.duration != chord.duration:
        #    from pdb import set_trace; set_trace()
    return chord(**new_parts)


def get_score_between(score, start=None, end=None):
    """Get the score with melodies that are between start and end

    Parameters
    ----------
    score :
        initial score
    start :
        Start time (relative to score) (Default value = None)
    end :
        End time (relative to score) (Default value = None)

    Returns
    -------
    type
        score, Score, modified score

    """
    start = start if start is not None else 0
    end = end if end is not None else score.duration
    new_score = None
    time = 0
    for chord in score.chords:
        chord_start = time
        chord_end = time + chord.duration
        if chord_end <= start:
            time += chord.duration
            continue
        elif chord_start >= end:
            # We are already arrived at the end of time
            break
        elif chord_end < end and chord_start >= start:
            # Perfect case we just copy the full chord
            new_score += chord.copy()
        else:
            # In all other cases we have to cut the chord
            new_start = start - time
            new_end = end - time
            new_chord = score.get_chord_between(chord, new_start, new_end)
            new_score += new_chord

        time += chord.duration

    return new_score



from .element import Formatter

class VoicingElement(Formatter):

    def __init__(self, notes):
        self.notes = notes

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        return f'!voicing {" ".join([str(n) for n in self.notes])}'


from .element import Formatter

class RhythmElement(Formatter):

    def __init__(self, instrument, rhythm):
        self.instrument = instrument
        self.rhythm = rhythm

    def to_text(self):
        return '\n' + str(self)

    def __repr__(self):
        return f'!rhythm {self.instrument} {self.rhythm}'



from musiclang import Melody, Score
from musiclang.transform import ScoreTransformer
from musiclang import Continuation

class ContinuationWhenSameNote(ScoreTransformer):
    """
    When a note is repeated create a continuation instead of repeating the note
    """

    def __init__(self, only_inside_chord=False, instruments=None, keep_instruments=None):
        self.only_inside_chord = only_inside_chord
        self.keep_instruments = keep_instruments if keep_instruments is not None else []
        self.instruments = instruments

    def action(self, score: Score, **kwargs):
        new_score = score.copy()
        instruments = new_score.instruments
        for inst in instruments:
            current_pitch = None
            for chord in new_score.chords:
                if self.only_inside_chord:
                    current_pitch = None
                if inst in chord.score and inst not in self.keep_instruments:
                    if self.instruments is None or inst in self.instruments:
                        mel = chord.score[inst]
                        for idx, note in enumerate(mel.notes):
                            if note.is_note:
                                pitch = chord.to_pitch(note)
                                if current_pitch == pitch:
                                    mel.notes[idx] = Continuation(note.duration)
                                current_pitch = pitch
                            elif note.is_continuation:
                                pass
                            else:
                                current_pitch = None

        return new_score




import numpy as np

"""
The algorithm is based on this paper :
HMM-Based Voice Separation of MIDI
Performance
Andrew McLeod
University of Edinburgh
Mark Steedman
University of Edinburgh
https://homepages.inf.ed.ac.uk/steedman/papers/music/VoiceSeparation.pdf

Tunable parameters : 
- sigp
- gmin
- sigg
- snew
"""

from .item import Item
import pickle

def separate_voices(notes):
    """Input a sequence of notes

    Parameters
    ----------
    notes :
        return: Sequence of note items with voice correctly
        set with the following algorithm : (https://homepages.inf.ed.ac.uk/steedman/papers/music/VoiceSeparation.pdf)

    Returns
    -------
    type
        Sequence of note items with voice correctly
        set with the following algorithm : (https://homepages.inf.ed.ac.uk/steedman/papers/music/VoiceSeparation.pdf)

    """

    # Split notes in groups that are playing simultaneously
    splitted_notes = split_notes(notes)
    voices = voice_separation_algorithm(splitted_notes)
    new_notes = []
    for idx, voice in enumerate(voices):
        for n in voice:
            n[6] = idx
            new_notes.append(n)

    return Item.frommatrix(new_notes)


def voice_separation_algorithm(splitted_notes, beam_size=1, max_states=1):
    """Implementation of the voice separation algorithm

    Parameters
    ----------
    splitted_notes :
        param beam_size: Number of best states to keep for each iteration
    max_states :
        Number of state to keep for each iteration (Default value = 1)
    beam_size :
         (Default value = 1)

    Returns
    -------

    """
    S = [([], 1)] # State, logprob of state
    for idx, incoming_notes in enumerate(splitted_notes):
        for n in incoming_notes:
            S = sum([evaluate_new_states(s, logprob, n, beam_size) for s, logprob in S], [])
            if len(S) > max_states:
                S = list(sorted(S, key=lambda x: -x[-1]))[:max_states]
    # Find max
    final_path, logprob = list(sorted(S, key=lambda x: -x[-1]))[0]
    return final_path

def evaluate_new_states(S, logprob, n, beam_size):
    """Evaluate all candidate states for new note n in state S

    Parameters
    ----------
    S :
        param logprob:
    n :
        param beam_size:
    logprob :

    beam_size :


    Returns
    -------

    """
    probas = generate_probas_for_one(S, n)
    assert len(probas) > 0
    # Save best 2
    probas = list(sorted(probas, key=lambda x: -x[1]))[:beam_size]
    # Save new_S, proba
    new_states = [(get_new_state(S, n, w), logprob + proba) for w, proba in probas]
    return new_states

def split_notes(notes):
    """Split notes in a group of notes that are playing simultaneously

    Parameters
    ----------
    notes :
        return: list[list[note]]

    Returns
    -------
    type
        list[list[note]]

    """
    notes = sorted(notes, key=lambda x: x[0])
    splitted_notes = []
    time = notes[0][0]
    current_split = []
    for note in notes:
        if note[0] != time:
            splitted_notes.append(sorted(current_split, key=lambda x: x[3]))
            current_split = []
            time = note[0]
        current_split.append(note)

    splitted_notes.append(sorted(current_split, key=lambda x: x[3]))
    return splitted_notes

def get_new_state(S, n, w):
    """

    Parameters
    ----------
    S :
        
    n :
        
    w :
        

    Returns
    -------

    """
    import copy
    import time
    new_S = pickle.loads(pickle.dumps(S, -1))
    if w > 0:
        new_S[w-1].append(n)
    elif w < 0:
        new_S.insert(-w-1, [n])
    else:
        raise Exception('w must not be 0')
    return new_S


def generate_probas_for_one(S, n):
    """

    Parameters
    ----------
    S :
        
    n :
        

    Returns
    -------

    """
    nb_voices = len(S)
    new_voices = np.arange(-1, -nb_voices-2, -1).tolist()
    existing_voices = np.arange(1, nb_voices+1).tolist()
    # Filter new_voices and existing voices
    W = new_voices + existing_voices

    # Filter possible movements
    W = [w for w in W if w < 0 or can_be_same_voice(S[w-1][-1], n)]
    # Filter only max orders

    # If can be assigned to existing voice, then bypass
    if len([w for w in W if w > 0]) > 0:
        W = [w for w in W if w > 0]
    orders = [order(S, n, w) for w in W]
    max_order = max(orders)
    W = [w for w, ord in zip(W, orders) if ord == max_order]

    probas = [logP_Ti(S, n, w) for w in W]

    return list(zip(W, probas))

def on(n):
    """On time of note

    Parameters
    ----------
    n :
        note

    Returns
    -------

    """
    return float(n[0])

def off(n):
    """Off time of note

    Parameters
    ----------
    n :
        note

    Returns
    -------

    """
    return float(n[1])

def duration(n):
    """Duration of note

    Parameters
    ----------
    n :
        return:

    Returns
    -------

    """
    return off(n) - on(n)

def num(n):
    """Pitch of note

    Parameters
    ----------
    n :
        return:

    Returns
    -------

    """
    return n[3]

def can_be_same_voice(n1, n2):
    """Can n1 and n2 be on the same voice
    There is a condition on the overlapping of the two notes

    Parameters
    ----------
    n1 :
        param n2:
    n2 :
        

    Returns
    -------

    """
    cond1 = off(n1) - on(n2) <= duration(n1)/2
    cond2 = off(n1) < off(n2)
    return cond1 and cond2

def gauss(mu, sig):
    """Gaussian density

    Parameters
    ----------
    mu :
        
    sig :
        

    Returns
    -------

    """
    return np.exp(-((mu) / sig)**2)


def g(S, n, w, sig_g=127000):
    """Estimate a penalty about the time gap between new note added in the specific voice of S

    Parameters
    ----------
    S :
        State
    n :
        note
    w :
        voice index of note
    sig_g :
        std of g, tunable parameter (Default value = 127000)

    Returns
    -------

    """
    last_V_w = S[w-1][-1]
    return np.log(1-(on(n) - off(last_V_w))/sig_g) + 1

def gap(S, n, w, gmin=9e-5):
    """Bound g gap function with a gmin

    Parameters
    ----------
    S :
        param n:
    w :
        param gmin:
    n :
        
    gmin :
         (Default value = 9e-5)

    Returns
    -------

    """
    return max(g(S, n, w), gmin)

# sigp = 4 before
def pitch(S, n, w, sigp=20):
    """Return the deviation of note pitch with average pitch of the state voice with gaussian density

    Parameters
    ----------
    S :
        State
    n :
        note
    w :
        index in state where to add the note
    sigp :
        standard deviation of the density (Default value = 20)

    Returns
    -------

    """
    return gauss(num(n) - average_pitch(S[w-1]), sigp)


def logP_Ti(S, ni, wi):
    """Log-probability of a state transition of ni in voice wi - 1 of S

    Parameters
    ----------
    S :
        State
    ni :
        note
    wi :
        index in state where to add the note

    Returns
    -------

    """
    return np.log(P(S, ni, wi) + 1e-13) + np.log(order(S, ni, wi))

def P_T(S, n, w):
    """Probability of a state transition of notes n in voices w of S

    Parameters
    ----------
    S :
        State
    ni :
        list[note]
    wi :
        list[index in state where to add the note]
    n :
        
    w :
        

    Returns
    -------

    """
    return np.product([P(S, ni, wi) * order(S, ni, wi) for ni, wi in zip(n, w)])

def order(S, n, w):
    """Penalty to avoid note crossing, see : https://homepages.inf.ed.ac.uk/steedman/papers/music/VoiceSeparation.pdf

    Parameters
    ----------
    S :
        param n:
    w :
        index in state where to add the note
    n :
        

    Returns
    -------

    """
    case1 = (abs(w) > 1) and (average_pitch(S[abs(w) - 2]) > num(n)) # Crossing
    case2 = (0 < w < len(S)) and (average_pitch(S[w]) < num(n))
    case3 = (-len(S) <= w < 0) and (average_pitch(S[abs(w)-1]) < num(n))
    return 2 ** -(case1 + case2 + case3)

def P(S, n, w, s_new=0):
    """Probability

    Parameters
    ----------
    S :
        State
    n :
        note
    w :
        index in state where to add the note
    s_new :
        Proba of creating a new voice (Default value = 0)

    Returns
    -------

    """
    if w > 0:
        return pitch(S, n, w) * gap(S, n, w)
    else:
        return s_new

def average_pitch(V, l=5):
    """Calculate the weighted average pitch of the voice over a window of size l

    Parameters
    ----------
    V :
        param l:
    l :
         (Default value = 5)

    Returns
    -------

    """
    min_l_V = min(l, len(V))
    len_V = len(V)
    numerator = sum([(2**i) * num(V[len_V - i - 1]) for i in range(min_l_V)])
    denominator = sum([(2**i) for i in range(min_l_V)])
    return numerator / denominator



import pytest
from musiclang.library import *

import musiclang as ml

def test_notes_equal():

    n1 = ml.Note("s", 1, 0, 1)
    n2 = ml.Note("s", 1, 0, 1)
    assert n1 == n2

def test_notes_octave_different_not_equal():
    n1 = ml.Note("s", 1, 0, 1)
    n2 = ml.Note("s", 1, 1, 1)
    assert n1 != n2

def test_notes_duration_different_not_equal():
    n1 = ml.Note("s", 1, 0, 1)
    n2 = ml.Note("s", 1, 0, 2)
    assert n1 != n2

def test_notes_type_different_not_equal():
    n1 = ml.Note("s", 1, 0, 1)
    n2 = ml.Note("su", 1, 0, 1)
    assert n1 != n2

def test_notes_val_different_not_equal():
    n1 = ml.Note("su", 2, 0, 1)
    n2 = ml.Note("su", 1, 0, 1)
    assert n1 != n2



def test_as_key():
    note = s4.o(1).e.fff
    assert note.as_key() == s4

    note = s4.o(1).e.fff
    assert note.as_key(octave=True) == s4.o(1)

    note = s4.o(1).e.fff
    assert note.as_key(amp=True) == s4.fff
    note = s4.o(1).e.fff

    assert note.as_key(duration=True) == s4.e


def test_to_extension_note():
    chord = (I % I.M)['6']
    note = s2
    assert note.to_extension_note(chord) == b0
    assert note.o(1).to_extension_note(chord) == b0.o(1)
    assert note.e.o(1).to_extension_note(chord) == b0.e.o(1)
    assert note.e.o(1).fff.to_extension_note(chord) == b0.e.o(1).fff

    chord = (I % I.M)['2']
    note = s6
    assert note.to_extension_note(chord) == b0.o(1)
    note = s5
    assert note.to_extension_note(chord) == s5

    chord = (I % I.M)['2[add6]']
    note = s5
    assert note.to_extension_note(chord) == b4


def test_to_chord_note():
    chord = (I % I.M)['6']
    note = s2
    assert note.to_chord_note(chord) == c1
    assert note.o(1).to_chord_note(chord) == c1.o(1)
    assert note.e.o(1).to_chord_note(chord) == c1.e.o(1)
    assert note.e.o(1).fff.to_chord_note(chord) == c1.e.o(1).fff
    chord = (I % I.M)['2[add6]']
    note = s5
    assert note.to_chord_note(chord) == c3



"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import pandas as pd
import numpy as np
from mido import MidiFile
import time
import warnings

TYPE = "type"
TIME = "time"
NOTE = "note"
VEL = "vel"
CHANNEL = "channel"
DURATION = "duration"
END_TIME = "end_time"
START_TIME = "start_time"
TRACK = "track"
VOICE = "voice"
ON = 1
OFF = 0


def parse_midi(filename, **kwargs):
    """Parse a midi files and returns array of notes, instruments and the tempo

    Parameters
    ----------
    filename :
        param kwargs:
    **kwargs :
        

    Returns
    -------

    """
    notes, config, bars = _load_midi(filename, **kwargs)
    instruments = config['instruments']
    tempo = config['tempo']
    time_signature = config['bar_duration']
    return notes, instruments, tempo, time_signature, bars


"""
Below, some loader helpers :
"""

class MusicLangIgnoreException(Exception):
    """ """
    pass


def _load_midi(filename, ignore_file_with_bar_change=False, **kwargs):
    """Create a note array from a midi file

    Parameters
    ----------
    filename :
        return:
    ignore_file_with_bar_change :
         (Default value = False)
    **kwargs :
        

    Returns
    -------

    """
    notes, config, bars = _parse(filename)
    if len(config['bar_durations']) > 1 and ignore_file_with_bar_change:
        raise MusicLangIgnoreException('Bar duration change events in midifile, MusicLang cannot parse that')
    elif len(config['bar_durations']) == 1:
        config['bar_duration'] = config['bar_durations'][0]
    else:
        config['bar_duration'] = None
    return notes, config, bars



def _parse(filename, **kwargs):
    """

    Parameters
    ----------
    filename :
        
    **kwargs :
        

    Returns
    -------

    """
    from fractions import Fraction as frac
    from .load_score import load_score
    mf = MidiFile(filename)
    instruments = _infer_instruments(mf)
    notes = []
    bar_durations = []
    tempos = []
    for track_idx, track in enumerate(mf.tracks):
        t = 0
        for note in track:
            if note.type == 'note_on':
                notes.append([1, t + note.time, note.note, note.velocity, note.channel, track_idx])
            elif note.type == 'note_off':
                notes.append([0, t + note.time, note.note, note.velocity, note.channel, track_idx])
            elif note.type == 'time_signature':
                bar_durations.append(note.numerator * frac(4, note.denominator))
            elif note.type == 'set_tempo':
                tempos.append((time, note.tempo))
            t = t + note.time
    first_tempo = int(60/(tempos[0][1]/1e6))
    config = {'ticks_per_beats': mf.ticks_per_beat,
             'instruments': instruments,
             'tempo': first_tempo, 'tempos': tempos, 'bar_durations': bar_durations}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        notes_score, bars = load_score(filename, ticks_per_beat=mf.ticks_per_beat)
    notes_score_df = pd.DataFrame(notes_score, columns=[START_TIME, END_TIME, VEL, NOTE, TRACK, CHANNEL, VOICE])
    notes_score_df[TRACK] = (notes_score_df[TRACK] + 1).astype(int)
    notes_score_df[DURATION] = notes_score_df[END_TIME] - notes_score_df[START_TIME]
    notes_score_df[VOICE] = (notes_score_df[VOICE] - 1).astype(int)

    if notes_score_df[START_TIME].min() < 0:
        # Add bar duration in notes_score[START_TIME]
        bar_duration = config['bar_durations'][0]
        notes_score_df[START_TIME] += bar_duration
        for idx in range(len(bars)):
            bars[idx] = bars[idx][0] + bar_duration, bars[idx][1] + bar_duration

    if len(bars) > 1 and (bars[-1][1] - bars[-1][0] != bars[-2][1] - bars[-2][0]):
        bars[-1] = bars[-1][0], bars[-1][0] + bars[-2][1] - bars[-2][0]
    #notes_score_df[START_TIME] = notes_score_df[START_TIME] - notes_score_df[START_TIME].min() # Start at 0
    notes = np.asarray(notes_score_df[[START_TIME, NOTE, DURATION, VEL, TRACK, CHANNEL, VOICE]])
    return notes, config, bars


def _get_notes_in_beats(df, ticks_per_beats):
    """By default notes are stored in ticks from midi file, convert it to beats
    :return:

    Parameters
    ----------
    df :
        
    ticks_per_beats :
        

    Returns
    -------

    """
    df = df.copy()
    df[START_TIME] /= ticks_per_beats
    df[END_TIME] /= ticks_per_beats
    df[DURATION] /= ticks_per_beats
    return df

def _get_notes_dataframe(notes):
    """Convert note_on, note_off events to a note dataframe
    :return:

    Parameters
    ----------
    notes :
        

    Returns
    -------

    """
    from itertools import product
    events = pd.DataFrame(notes, columns=[TYPE, TIME, NOTE, VEL, CHANNEL, TRACK])
    times = {(track, note): 0 for track, note in product(events[TRACK].unique(), range(120))}
    vels = {(track, note): 0 for track, note in product(events[TRACK].unique(), range(120))}
    ons = {(track, note): False for track, note in product(events[TRACK].unique(), range(120))}
    notes = []
    for idx, row in events.iterrows():
        if (row[TYPE] == ON) and row[VEL] > 0 and not ons[row[TRACK], row[NOTE]]:
            times[row[TRACK], row[NOTE]] = row[TIME]
            vels[row[TRACK], row[NOTE]] = row[VEL]
            ons[row[TRACK], row[NOTE]] = True
        else: # Note OFF
            start_time = times[row[TRACK], row[NOTE]]
            ons[row[TRACK], row[NOTE]] = False
            vel = vels[row[TRACK], row[NOTE]]
            end_time = row[TIME]
            duration = end_time - start_time
            notes.append([start_time, end_time, duration, row[NOTE], vel, row[CHANNEL], row[TRACK], 0])

    notes = pd.DataFrame(notes, columns=[START_TIME, END_TIME, DURATION, NOTE, VEL, CHANNEL, TRACK, VOICE]).sort_values(START_TIME)
    return notes


def _infer_instruments(mf):
    """

    Parameters
    ----------
    mf :
        

    Returns
    -------

    """
    from musiclang.write.out.constants import REVERSE_INSTRUMENT_DICT
    channel_inst = {}
    for track in mf.tracks:
        for note in track:
            if note.type == 'program_change':
                if note.channel == 9:
                    channel_inst[note.channel] = f'drums_{note.program}'
                else:
                    channel_inst[note.channel] = REVERSE_INSTRUMENT_DICT[note.program]
    return channel_inst








from .composer import Composer

from musiclang import Tonality, ScoreFormatter, Element, Chord, Note, Score
from musiclang.transform.library import VoiceLeading
from musiclang.transform.composing.pattern import Pattern
from musiclang.library import NC

import numpy as np

def apply_pattern(pattern, chords, instruments=None, voicing=None, restart_each_chord=False,
                  fixed_bass=True, voice_leading=True, amp='mf'):
    """
    Apply a pattern to a chord progression
    Parameters
    ----------
    pattern: Score or Chord
        Pattern to use
    chords: Score or list
        Chord progression to use
    instruments: list[str] or None
        Instruments to use, by default use only piano
    voicing: list[Note] or None
        Voicing to use in the pattern. By default, spread the voicing between root -2 octave and fifth +1 octave
    restart_each_chord: bool
        If True, the pattern will be restarted at each chord
    fixed_bass: bool
        If True, the bass will be fixed in the voice leading
    voice_leading: bool
        If True, the voice leading optimizer will be applied
    amp: str
        Amplitude of the pattern (in ppp, pp, p, mp, mf, f, ff, fff)

    Returns
    -------
    Score

    """

    orchestra = {'bar_duration': pattern.duration, 'nb_instruments': len(pattern.instruments), 'pattern': pattern}

    if isinstance(chords, (list, tuple)):
        chords = sum(chords, None)
    if isinstance(chords, Chord):
        chords = Score(chords)
    if instruments is None:
        instruments = [f'piano' for i in range(orchestra['nb_instruments'])]

    # Clean instruments
    real_instruments = _clean_instrument_names(instruments)

    if voicing is None:
        values = np.linspace(-6, 5, orchestra['nb_instruments']).astype(int)
        voicing = [Note('b', value % 3, value // 3, 1) for value in values]

    patternator = {'restart_each_chord': restart_each_chord}
    return auto_compose([], chords, orchestra, voicing, patternator,
                 "C", [], real_instruments, acc_amp=amp,
                 fixed_bass=fixed_bass, voice_leading=voice_leading
                 )

def _clean_instrument_names(instruments):
    """
    Given a list of raw instrument, spread the voice indexes accordingly
    eg : ['piano', 'piano'] -> ['piano__0', 'piano__1']
    Returns
    -------
    list[str]
    """
    real_instruments = []
    instrument_dict = {}
    for instrument in instruments:
        clean_instrument = ''.join(instrument.split('__')[:-1])
        instrument_dict[clean_instrument] = instrument_dict.get(clean_instrument, -1) + 1
        index = instrument_dict[clean_instrument]
        real_instruments.append(instrument + f'__{index}')

    return real_instruments

def auto_compose(melody, harmony, orchestra, voicing, patternator,
                 tonality, solo_instrument, instruments, acc_amp='mf',
                 fixed_bass=True, voice_leading=True
                 ):
    from musiclang.transform import Patternator


    if isinstance(melody, list):
        melodies = melody
    else:
        melodies = [melody]

    if isinstance(voice_leading, list):
        voice_leadings = voice_leading
    else:
        voice_leadings = [voice_leading]

    if isinstance(orchestra, list):
        orchestras = orchestra
    else:
        orchestras = [orchestra]

    if isinstance(voicing, list) and not isinstance(voicing[0], list):
        voicings = [voicing]
    else:
        voicings = voicing

    if isinstance(solo_instrument, list):
        solo_instruments = solo_instrument
    else:
        solo_instruments = [solo_instrument]

    if isinstance(instruments, list) and not isinstance(instruments[0], list):
        instrumentss = [instruments]
    else:
        instrumentss = instruments

    if isinstance(fixed_bass, list):
        fixed_basses = fixed_bass
    else:
        fixed_basses = [fixed_bass for i in range(len(instrumentss))]

    real_tonality = Element(0) % Tonality.from_str(tonality)
    basses = [f'v_{j}__0' for j in range(len(instrumentss))]
    temp_instrumentss = [[f'v_{j}__{i}' for i in range(len(instr))] for j, instr in enumerate(instrumentss)]

    if len(melodies) > 0:
        score_theme = real_tonality(**{f'm__{idx}': melody for idx, melody in enumerate(melodies)}).to_absolute_note()
    else:
        score_theme = None

    flat_temp_instruments = sum(temp_instrumentss, [])
    flat_voicings = sum(voicings, [])
    if isinstance(harmony, str):
        chords = ScoreFormatter(harmony, instruments=flat_temp_instruments, voicing=flat_voicings).parse()
    else:
        chords = sum([chord(**{ins: voicing.set_duration(chord.duration)
                               for ins, voicing in zip(flat_temp_instruments, flat_voicings)}) for chord in harmony.to_score().chords], None)

    if score_theme is not None:
        score = score_theme.project_on_score(chords, keep_pitch=True, voice_leading=True, keep_score=True)
        assert score.duration == score_theme.duration, f"Durations don't match between theme ({score_theme.duration}) and chords"
    else:
        score = chords


    fixed_voices = [f'm__{i}' for i in range(len(melodies))]
    if fixed_bass:
        fixed_voices = fixed_voices + [bass for bass, fixed_bass in zip(basses, fixed_basses) if fixed_bass]
    for instruments, voice_leading in zip(temp_instrumentss, voice_leadings):
        if not voice_leading:
            fixed_voices = fixed_voices + instruments[1:]

    score = VoiceLeading(fixed_voices=fixed_voices)(score, max_norm=2, max_norm_rules=2)

    patternator = Patternator(**patternator)
    for orchestra, temp_instruments in zip(orchestras, temp_instrumentss):
        realized_orchestra = NC(**Pattern.from_json(orchestra, instruments=temp_instruments))
        score = patternator(realized_orchestra.set_amp(acc_amp), score)

    score = score.replace_instruments(**{temp_instr: instr for temp_instruments, instruments in zip(temp_instrumentss, instrumentss)
                                         for temp_instr, instr in zip(temp_instruments, instruments)})
    score = score.replace_instruments(**{f'm__{i}': solo_instrument for i, solo_instrument in enumerate(solo_instruments)})
    return score


class AutoComposer:
    """
    Class for composing a piece automatically using an annotation generator model from musiclang_predict
    """

    def __init__(self, model, parts, tonalities, form, form_orchestra, orchestras, signature=(4, 4), **kwargs):
        self.model = model
        self.parts = parts
        self.tonalities = tonalities
        self.form = form
        self.form_orchestra = form_orchestra
        self.signature = signature
        self.orchestras = orchestras
        self.kwargs = kwargs

    def predict(self, signature, tonality, temperature=1.0, max_new_tokens=500, seed=None, **kwargs):
        start = f"Time Signature: {signature[0]}/{signature[1]}\nm0 {tonality}:"
        return self.model.predict(start=start,
                             temperature=temperature,
                             max_new_tokens=max_new_tokens,
                             seed=seed)

    def compose(self, **kwargs):
        """
        Main method to compose a song automatically with the provided material

        Parameters
        ----------
        kwargs

        Returns
        -------
        score: Score
        text: str
            Final annotation predicted by the auto composer

        """
        progressions = self.get_progressions(**kwargs)
        composer = Composer(progressions,
                            self.form,
                            self.form_orchestra,
                            self.orchestras,
                            signature=self.signature,
                            **self.kwargs)
        return composer.compose()

    def get_progressions(self, **kwargs):
        progressions = []
        for idx, (nb_bars, tonality) in enumerate(zip(self.parts, self.tonalities)):
            progression = self.predict(self.signature, tonality, **kwargs)
            progression = self.get_n_bars(nb_bars, progression)
            progressions.append(progression)
        return progressions

    def line_is_bar(self, line):
        return line.lstrip(' ').startswith('m')

    def get_n_bars(self, n, progression):
        lines = progression.split('\n')
        bars = [line for line in lines if self.line_is_bar(line)]
        new_lines = []
        nb_bars = 0
        for line in lines:
            if self.line_is_bar(line):
                new_lines.append(bars[nb_bars % len(bars)])
                nb_bars += 1
            else:
                new_lines.append(line)
            if nb_bars == n:
                break
        if nb_bars < n:
            for i in range(nb_bars, n):
                new_lines.append(bars[(nb_bars + i) % len(bars)])

        return '\n'.join(new_lines)



from musiclang import Tonality


def test_tonalities_equal():

    t1 = Tonality(0,  mode="m", octave=0)
    t2 = Tonality(0,  mode="m", octave=0)
    assert t1 == t2
    assert t1 == t1.copy()


def test_tonalities_degree_not_equal():

    t1 = Tonality(0,  mode="m", octave=0)
    t2 = Tonality(1, mode="m", octave=0)
    assert t1 != t2


def test_abs_degree():
    t1 = Tonality(2, mode="m", octave=0)
    assert t1.abs_degree == 2


def test_tonalities_mode_not_equal():

    t1 = Tonality(0,  mode="m", octave=0)
    t2 = Tonality(1,  mode="M", octave=0)
    assert t1 != t2

def test_tonalities_octave_not_equal():

    t1 = Tonality(0,  mode="m", octave=0)
    t2 = Tonality(1,  mode="m", octave=1)
    assert t1 != t2



def test_tonality_addition_octave():
    t1 = Tonality(5,  mode="M", octave=0)
    t2 = Tonality(6,  mode="m", octave=0)
    result = t1 + t2
    assert result == Tonality(11, mode="m", octave=0)

def test_enharmonic_equalities():
    # #Fifths == fifths
    assert Tonality(4, mode="m", octave=1) == Tonality(4,  mode="m", octave=1)


def test_tonality_addition_thirds():
    # For simplification : III# of C is considered as E not E#, idem for VII#
    t1 = Tonality(2, mode="M", octave=1)
    t2 = Tonality(1,  mode="m", octave=0)
    result = t1 + t2
    assert result == Tonality(3, mode="m", octave=1)


def test_tonality_addition_with_accidents():
    # For simplification : III# of C is considered as E not E#, idem for VII#
    t1 = Tonality(2, mode="m", octave=1)
    t2 = Tonality(2, mode="m", octave=0)
    result = t1 + t2
    assert result == Tonality(4, mode="m", octave=1)


def test_tonality_to_scale_pitches():
    t1 = Tonality(1, mode="m", octave=1)
    scale_pitches = t1.scale_pitches

    assert scale_pitches == [13, 15, 16, 18, 20, 21, 24]


def test_abs_degree_of_mode():
    T = Tonality(1, mode="locrian", octave=0)
    assert T.abs_degree == 1

def test_abs_degree_of_sum_of_mode():
    T = Tonality(1, mode="locrian", octave=0)
    assert (T + T).abs_degree == 2

def test_tonality_to_scale_pitches_mode():
    t1 = Tonality(0,  mode="dorian", octave=0)
    scale_pitches = t1.scale_pitches

    assert scale_pitches == [0, 2, 3, 5, 7, 9, 10]


# "", b, n, s : b force to be minor interval, n : major interval, "" normal mode interval
# 3s.M = E

def test_change_mode():
    t1 = Tonality(2,  mode="m", octave=0)
    new_mode = "locrian"
    t2 = t1.change_mode(new_mode)
    assert t2.scale_pitches[0] == t1.scale_pitches[0]


from .utils_random import *
from .mask import Mask


class Transformer:
    TYPE = 'TRANSFORMER'

    def __call__(self, element, on=Mask(), **kwargs):
        raise NotImplemented

    def action(self, element, **kwargs):
        raise NotImplemented

    def get_params(self):
        dic = self.__dict__
        dic = {key: value for key, value in dic.items() if not key.startswith('_')}
        return dic

    def set_params(self, **params):
        for param in params.keys():
            self.__setattr__(param, params[param])

    def get_default(self, element):
        return element.copy()

    def copy(self):
        obj = self.__class__(**self.get_params())
        return obj

    def is_valid(self):
        return self.get_params()

    def init_seed(self, **kwargs):
        seed = None
        if "seed" in kwargs.keys():
            seed = kwargs['seed']
        self.seed = seed if seed is not None else np.random.randint(0, 2 ** 31 - 1)
        return self.seed

    def __repr__(self):
        params = self.get_params()
        param_str = "("
        for name, value in params.items():
            if isinstance(value, str):
                value = '"' + value + '"'
            param_str += str(name) + "=" + str(value) + ","

        param_str = param_str[:-1]
        param_str += ")"
        return self.__class__.__name__ + param_str

    def get_part_name(self, part, chord):
        if isinstance(self.part, int):
            return chord.parts[self.part]
        else:
            return part

    def get_parts_name(self, parts, chord):
        return [self.get_part_name(part, chord) for part in parts]

    def convert_to_list(self, val, n, default, constr, increase=True):
        """
        Convert a parameter value to a list given a number
        If
        :param val:
        :param n:
        :param default:
        :param constr:
        :return:
        """
        if val is None:
            res = [default() for i in range(n)]
        elif isinstance(val, constr()):
            res = [val + 5000 * i * increase for i in range(n)]
        else:
            res = val

        return res

    @classmethod
    def random_choice(self, array, chars, seed="", root=""):
        """
        Select an element at random in an array
        """
        return random_choice(array, chars, seed=seed, root=root)

    @classmethod
    def random_permutation(self, chars, array, seed="", root=""):
        """
        Generate a permutation of size n following a bytes encoding
        if chars,seed,root,proba is the same it always generates the same output
        Algorithm : Fisher Yates shuffle  : https://en.wikipedia.org/wiki/Random_permutation#:~:text=A%20simple%20algorithm%20to%20generate,has%20index%200%2C%20and%20the
        :param array: array to shuffle
        :param seed:
        :param root:
        :return:
        """
        return random_permutation(chars, array, seed=seed, root=root)

    @classmethod
    def random_boolean(self, chars, seed="", root="", proba=0.2):
        """
        Generate a random boolean following a bytes encoding
        if chars,seed,root,proba is the same it always generates the same output
        :param seed:
        :param root:
        :param chars:
        :param proba:
        :return:
        """
        return random_boolean(chars, seed=seed, root=root, proba=proba)

    @classmethod
    def random_int(self, chars, low, high, seed="", root=""):
        """
        Generate an integer between low and high following a bytes encoding
        if chars,seed,root,proba is the same it always generates the same output
        :param chars:
        :param low:
        :param high:
        :param seed:
        :param root:
        :return:
        """
        return random_int(chars, low, high, seed=seed, root=root)

    @classmethod
    def random_float(self, chars, low, high, seed="", root=""):
        """
        Generate a float between low and high following a bytes encoding
        if chars,seed,root,proba is the same it always generates the same output
        :param chars:
        :param low:
        :param high:
        :param seed:
        :param root:
        :return:
        """
        return random_float(chars, low, high, seed=seed, root=root)

    def __add__(self, other):
        if isinstance(other, Transformer):
            return TransformerSerie([self, other])
        if isinstance(other, TransformerSerie):
            return TransformerSerie([self] + other.actions)
        if isinstance(other, TransformerParallel):
            return TransformerSerie([self, other])
        else:
            raise Exception(f'Cannot add Transformer {type(self)} and {type(other)}')

    @staticmethod
    def get_instrument_name_and_part(instrument_name):
        """
        Return intrument_name, instrument (before __) and part (after __) as int
        :param instrument_name:
        :return:
        """
        parts = instrument_name.split('__')
        if len(parts) == 1:
            inst, part = parts[0], 0
        else:
            inst, part = parts
            part = int(part)
        return instrument_name, inst, part

    @staticmethod
    def part_is_of_instrument(part, instrument):
        return part.split('__')[0] == instrument

    @staticmethod
    def get_instrument_parts(inst, chord: 'Chord'):
        return [part for part in chord.score.keys() if Transformer.part_is_of_instrument(part, inst)]

    @staticmethod
    def get_instrument_parts_separated(inst, chord: 'Chord'):
        return [part.split('__') for part in chord.score.keys() if Transformer.part_is_of_instrument(part, inst)]

    def apply_on_melody(self, element, on=Mask(), **kwargs):
        from musiclang import Melody
        beat = 0
        idx = 0
        last_note = None
        new_melody = []
        for m in element.notes:
            note = self.action(m, beat=beat, idx=idx,
                               last_note=last_note, **kwargs) if on(m, beat=beat, idx=idx,
                                                                    last_note=last_note,
                                                                    **kwargs) else self.get_default(m)
            beat += m.duration
            idx += 1
            last_note = m
            if note is not None:
                new_melody.append(note)

        new_melody = [m for m in new_melody if m is not None]
        return Melody(new_melody, tags=element.tags)

    def apply_on_chord(self, element, on=Mask(), **kwargs):
        from musiclang import Chord
        chord = Chord(element.element, extension=element.extension, tonality=element.tonality,
                      score={
                          key: self(element.score[key], on=on.child(element, **kwargs), chord=element, instrument=key, **kwargs)
                          if on(element.score[key], chord=element, instrument=key, **kwargs)
                          else self.get_default(element.score[key]) for key in element.score},
                      octave=element.octave, tags=element.tags)
        # Filter None keys
        return chord(**{part: melody for part, melody in chord.score.items() if melody is not None})

    def apply_on_score(self, element, on=Mask(), **kwargs):
        beat = 0
        idx = 0
        last_chord = None
        score = None
        for m in element.chords:
            chord = self(m, on=on.child(element, **kwargs), chord_beat=beat, chord_idx=idx, last_chord=last_chord, **kwargs)\
                if on(m, chord_beat=beat, chord_idx=idx, last_chord=last_chord, **kwargs) else self.get_default(m)
            beat += m.duration
            idx += 1
            last_chord = m
            if chord is not None:
                score += chord

        return score.add_tags(element.tags)

    @staticmethod
    def get_part(inst, voice, chord: 'Chord'):
        return chord.score.get(inst + '__' + voice, None)





from .project import get_melody_between, project_on_rhythm, get_absolute_voice, get_nearest_note_in_context

def arrange_melody_with_chords(melody, patterns, score, inst):
    """

    Parameters
    ----------
    melody :
        
    patterns :
        
    score :
        
    inst :
        

    Returns
    -------

    """
    time = 0
    new_score = None
    for pattern, chord in zip(patterns, score.chords):
        submelody = get_melody_between(melody, time, time + pattern.duration)
        submelody = submelody & - chord.scale_degree
        submelody = modify_melody_to_be_on_chord(submelody, pattern, chord)
        new_chord = chord.copy()(**{**chord.score, inst: submelody})
        new_score += new_chord
        time += pattern.duration
        assert submelody.duration == pattern.duration, "Wrong duration {} {}".format(submelody.duration, pattern.duration)
        assert new_chord.duration == pattern.duration
    assert new_score.duration == sum([p.duration for p in patterns])
    return new_score


def modify_melody_to_be_on_chord(melody, rythm, chord, dict_notes=None):
    """Modify the melody to be on notes specified

    Parameters
    ----------
    rythm :
        return:
    melody :
        
    chord :
        
    dict_notes :
         (Default value = None)

    Returns
    -------

    """

    melody = get_absolute_voice(melody)
    if dict_notes is not None:
        assert all([d.duration == 1 for d in dict_notes.keys()]), "Dict notes should be quarter notes in this method"
    new_melody = None

    projected_rythm = project_on_rhythm(melody, rythm)
    for note, rythm_note in zip(melody.notes, projected_rythm.notes):
        # Get note playing at
        # get nearest dict_notes
        rythm_note = rythm_note.set_duration(1)
        new_melody += get_nearest_note_in_context(rythm_note, note, chord, dict_notes=dict_notes)

    return new_melody








from musiclang.transform.library import ReverseMelody
from musiclang.library import *

def test_reverse_melody_on_chord():

    reverser = ReverseMelody()
    score = (I % I.M)(s0 + s1 + s2)

    result = reverser(score)
    expected_score = (I % I.M)(s2 + s1 + s0)

    assert result == expected_score



def test_reverse_melody_on_score():

    reverser = ReverseMelody()
    score = (I % I.M)(s0 + s1 + s2) + (V % I.M)(violin__1=s0 + s1)

    result = reverser(score)
    expected_score = (I % I.M)(s2 + s1 + s0) + (V % I.M)(violin__1=s1 + s0)

    assert result == expected_score


def test_reverse_melody_on_melody():

    reverser = ReverseMelody()
    melody = s1 + s2.h + s3
    result = reverser(melody)
    expected_melody = s3 + s2.h + s1

    assert result == expected_melody


from .element import Formatter
from .voice_leading import VoiceLeadingElement
from .instruments import InstrumentsElement
from .counterpoint import CounterpointElement
from .rhythm import RhythmElement
from .time_signature import TimeSignatureElement
from .tonality import TonalityElement
from .chord import ChordElement
from .voicing import VoicingElement
from .free_text import FreeTextElement
from .bar import BarElement
from .beat import BeatElement


from musiclang.transform import ScoreTransformer


class RepeatScore(ScoreTransformer):
    def __init__(self, n):
        self.n = n

    def action(self, score, **kwargs):
        return score * self.n



class Mask:
    """
    Class responsible to make a masked query on a score.
    A Mask can check if a query is true on a given element (score, chord, note, melody)
    and give a subquery that can be used on a child of the element (Mask.child).
    It is used internally by the Transform, Filter classes to either apply a transformation
    on a filter of a score or filter a score directly.

    Examples
    --------

    A mask that checks if a chord has a tag "label"
    >>> from musiclang.library import *
    >>> from musiclang.transform import Mask
    >>> mask = Mask.Chord() > Mask.Has("label")
    >>> mask((I % I.M).add_tag('label'))
    True

    """
    def __call__(self, element, **kwargs):
        return True

    @classmethod
    def eval(cls, query):
        # Add to the namespace
        has = Mask.Has
        has_at_least = Mask.HasAtLeast
        instruments = Mask.InstrumentIn
        Score = Mask.Score()
        Melody = Mask.Melody()
        Note = Mask.Note()
        Chord = Mask.Chord()
        true = Mask.True_()
        false = Mask.False_()
        query = eval(query)
        return query

    def child(self, element, **kwargs):
        return self

    # Note masks

    @classmethod
    def NoteIn(cls, notes, ignore_octave=False, ignore_rythm=False):
        return cls.Note() > NoteInMask(notes, ignore_octave=ignore_octave, ignore_rythm=ignore_rythm)

    @classmethod
    def BeatIn(cls, beats):
        return cls.Note() > BeatInMask(beats)

    @classmethod
    def BeatPlayingIn(cls, beats):
        return cls.Note() > BeatPlayingInMask(beats)

    @classmethod
    def DurationIn(cls, durations):
        return cls.Note() > DurationInMask(durations)


    @classmethod
    def BeatBetween(cls, start, end):
        return cls.Note() > BeatBetweenMask(start, end)


    @classmethod
    def DurationBetween(cls, start, end):
        return cls.Note() > DurationBetweenMask(start, end)

    # Melodic masks

    @classmethod
    def InstrumentIn(cls, instruments):
        return cls.Melody() > InstrumentsMask(instruments)

    # Chord And Tonality masks

    @classmethod
    def ChordBeatIn(cls, beats):
        return cls.Chord() > ChordBeatInMask(beats)

    @classmethod
    def ChordBeatPlayingIn(cls, beats):
        return cls.Chord() > ChordBeatPlayingInMask(beats)

    @classmethod
    def ChordDurationIn(cls, durations):
        return cls.Chord() > ChordDurationInMask(durations)

    @classmethod
    def ChordBeatBetween(cls, start, end):
        return cls.Chord() > ChordBeatBetweenMask(start, end)

    @classmethod
    def ChordDurationBetween(cls, start, end):
        return cls.Chord() > ChordDurationBetweenMask(start, end)

    @classmethod
    def ModeIn(cls, modes):
        return cls.Chord() > ModeInMask(modes)

    @classmethod
    def ChordDegreeIn(cls, chord_degrees):
        return cls.Chord() > ChordDegreeInMask(chord_degrees)

    @classmethod
    def ChordExtensionIn(cls, chord_degrees):
        return cls.Chord() > ChordExtensionInMask(chord_degrees)

    @classmethod
    def TonalityDegreeIn(cls, tonality_degrees):
        return cls.Chord() > TonalityDegreeInMask(tonality_degrees)

    @classmethod
    def ChordIn(cls, chords, ignore_octave=False):
        return cls.Chord() > ChordInMask(chords, ignore_octave=ignore_octave)

    @classmethod
    def TonalityIn(cls, tonalities, ignore_octave=False):
        return cls.Chord() > TonalityInMask(tonalities, ignore_octave=ignore_octave)


    # First order logic masks

    @classmethod
    def HasAtLeast(cls, tags):
        """
        Check if an element has at least these tags

        Parameters
        ----------
        tags: iterable[str] or str

        """
        return HasAtLeastMask(tags)


    @classmethod
    def OriginateFrom(cls, tags):
        """
        Check if a chord originate from a set of given pipeline step or one pipeline step

        Parameters
        ----------
        tags: iterable[str] or str

        """

        return cls.Chord() > HasMask([f'step_{tag}' for tag in tags])

    @classmethod
    def Has(cls, tags):
        """
        Check if an element has these tags (all of them)

        Parameters
        ----------
        tags: iterable[str] or str

        """

        return HasMask(tags)

    @classmethod
    def And(cls, masks):
        """
        Logical and between masks. You can also use the "&" python operator between masks

        Parameters
        ----------
        masks: iterable[Mask]

        """

        return AndMask(masks)

    @classmethod
    def Bool(cls, bool):
        """
        A mask that always returns bool when called

        Parameters
        ----------
        bool: boolean

        """

        return BoolMask(bool)

    @classmethod
    def True_(cls):
        """
        A mask that always returns True when called
        """

        return BoolMask(True)

    @classmethod
    def False_(cls):
        """
        A mask that always returns False when called
        """
        return BoolMask(False)

    @classmethod
    def Or(cls, tags):
        """
        Logical or between masks. You can also use the "&" python operator between masks

        Parameters
        ----------
        masks: iterable[Mask]

        """
        return OrMask(tags)

    @classmethod
    def Score(cls):
        """
        A Mask that will only apply to score. Usually used with another mask with the ">" operator

        Examples
        --------
        >>> from musiclang.transform import Mask
        >>> mask = Mask.Score() > Mask.Has("label")
        """

        return ScoreMask()

    @classmethod
    def Melody(cls):
        """
        A Mask that will only apply to a melody. Usually used with another mask with the ">" operator

        Examples
        --------
        Check if tag label in this melody :

        >>> from musiclang.transform import Mask
        >>> mask = Mask.Melody() > Mask.Has("label")
        """
        return MelodyMask()

    @classmethod
    def Note(cls):
        """
        A Mask that will only apply to a note. Usually used with another mask with the ">" operator

        Examples
        --------
        Check if tag label in this note :

        >>> from musiclang.transform import Mask
        >>> mask = Mask.Note() > Mask.Has("label")
        """

        return NoteMask()

    @classmethod
    def Chord(cls):
        """
        A Mask that will only apply to a chord. Usually used with another mask with the ">" operator

        Examples
        --------
        Check if tag label in this Chord :

        >>> from musiclang.transform import Mask
        >>> mask = Mask.Chord() > Mask.Has("label")
        """

        return ChordMask()

    @classmethod
    def Func(cls, f):
        """
        A mask that will apply a custom function. The function should have signature (element, **kwargs) -> bool
        """

        return FuncMask(f)

    def __and__(self, other):
        return AndMask([self, other])

    def __or__(self, other):
        return OrMask([self, other])

    def __invert__(self):
        return NotMask(self)

    def __repr__(self):
        return 'true'



class NoteInMask(Mask):

    def __init__(self, notes, ignore_octave=False, ignore_rythm=False):
        from musiclang import Note
        if isinstance(notes, Note):
            notes = {notes}

        self.notes = set(notes)
        self.ignore_octave = ignore_octave
        self.ignore_rythm = ignore_rythm
        if self.ignore_octave:
            self.notes = [c.o(-c.octave) for c in self.notes]
        if self.ignore_rythm:
            self.notes = [c.set_duration(1) for c in self.notes]

    def __call__(self, note, **kwargs):
        n = note
        if self.ignore_octave:
            n = n.o(-note.octave)
        if self.ignore_rythm:
            n = n.set_duration(1)
        return n in self.notes

    def __repr__(self):
        return f'note_in({self.notes}, {self.ignore_octave}, {self.ignore_rythm})'


class BeatInMask(Mask):

    def __init__(self, beats):
        from fractions import Fraction
        if isinstance(beats, (int, Fraction)):
            beats = {beats}
        self.beats = set(beats)

    def __call__(self, note, beat=0, **kwargs):
        return beat in self.beats

    def __repr__(self):
        return f'beat_in({self.beats})'

class BeatPlayingInMask(Mask):

    def __init__(self, beats):
        from fractions import Fraction
        if isinstance(beats, (int, Fraction)):
            beats = {beats}
        self.beats = set(beats)

    def __call__(self, note, beat=0, **kwargs):
        end_beat = beat + note.duration
        return any([beat <= b < end_beat for b in self.beats])

    def __repr__(self):
        return f'beat_in({self.beats})'

class DurationInMask(Mask):

    def __init__(self, durations):
        from fractions import Fraction
        if isinstance(durations, (int, Fraction)):
            durations = {durations}
        self.durations = set(durations)

    def __call__(self, note, beat=0, **kwargs):
        return note.duration in self.durations

    def __repr__(self):
        return f'duration_in({self.durations})'


class DurationBetweenMask(Mask):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, note, beat=0, **kwargs):
        return self.start <= note.duration < self.end

    def __repr__(self):
        return f'duration_between({self.start}, {self.end})'


class BeatBetweenMask(Mask):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, note, beat=0, **kwargs):
        return self.start <= beat < self.end

    def __repr__(self):
        return f'beat_between({self.start}, {self.end})'


# Melodic



class InstrumentsMask(Mask):

    def __init__(self, instruments):
        if isinstance(instruments, str):
            instruments = {instruments}

        self.instruments = set(instruments)

    def __call__(self, element, chord=None, instrument=None, **kwargs):
        return instrument in self.instruments

    def __repr__(self):
        return f"instruments({self.instruments})"


# Chords


class ChordBeatInMask(Mask):

    def __init__(self, beats):
        from fractions import Fraction
        if isinstance(beats, (int, Fraction)):
            beats = {beats}
        self.beats = set(beats)

    def __call__(self, chord, chord_beat=0, **kwargs):
        return chord_beat in self.beats

    def __repr__(self):
        return f'chord_beat_in({self.beats})'


class ChordBeatPlayingInMask(Mask):

    def __init__(self, beats):
        from fractions import Fraction
        if isinstance(beats, (int, Fraction)):
            beats = {beats}
        self.beats = set(beats)

    def __call__(self, chord, chord_beat=0, **kwargs):
        end_beat = chord_beat + chord.duration
        return any([chord_beat <= b < end_beat for b in self.beats])

    def __repr__(self):
        return f'chord_beat_in({self.beats})'


class ChordDurationInMask(Mask):

    def __init__(self, durations):
        from fractions import Fraction
        if isinstance(durations, (int, Fraction)):
            durations = {durations}
        self.durations = set(durations)

    def __call__(self, chord, chord_beat=0, **kwargs):
        return chord.duration in self.durations

    def __repr__(self):
        return f'chord_duration_in({self.durations})'


class ChordDurationBetweenMask(Mask):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, chord, chord_beat=0, **kwargs):
        return self.start <= chord.duration < self.end

    def __repr__(self):
        return f'chord_duration_between({self.start}, {self.end})'


class ChordBeatBetweenMask(Mask):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, chord, chord_beat=0, **kwargs):
        return self.start <= chord_beat < self.end

    def __repr__(self):
        return f'chord_beat_between({self.start}, {self.end})'


class ModeInMask(Mask):

    def __init__(self, modes):
        if isinstance(modes, str):
            modes = {modes}
        self.modes = set(modes)

    def __call__(self, element, **kwargs):
        return element.tonality.mode in self.modes

    def __repr__(self):
        return f"mode_in({self.modes})"


class ChordInMask(Mask):

    def __init__(self, chords, ignore_octave=False):
        from musiclang import Chord
        if isinstance(chords, Chord):
            chords = {chords}
        self.chords = set(chords)
        self.ignore_octave = ignore_octave
        if self.ignore_octave:
            self.chords = [c.o(-c.octave) for c in self.chords]

    def __call__(self, element, **kwargs):
        if self.ignore_octave:
            return element.o(-element.octave) in self.chords
        else:
            return element in self.chords

    def __repr__(self):
        return f"chord_in({self.chords},ignore_octave={self.ignore_octave})"


class TonalityInMask(Mask):

    def __init__(self, tonalities, ignore_octave=False):
        from musiclang import Tonality
        if isinstance(tonalities, Tonality):
            tonalities = {tonalities}
        self.tonalities = set(tonalities)
        self.ignore_octave = ignore_octave
        if self.ignore_octave:
            self.tonalities = [c.o(-c.octave) for c in self.tonalities]

    def __call__(self, element, **kwargs):
        if self.ignore_octave:
            return element.tonality.o(element.tonality.octave) in self.tonalities
        else:
            return element.tonality in self.tonalities

    def __repr__(self):
        return f"tonality_in({self.tonalities},ignore_octave={self.ignore_octave})"


class ChordDegreeInMask(Mask):

    def __init__(self, degrees):
        if isinstance(degrees, int):
            degrees = {degrees}
        self.degrees = set(degrees)

    def __call__(self, element, **kwargs):
        return element.degree in self.degrees

    def __repr__(self):
        return f"chord_degree_in({self.degrees})"


class ChordExtensionInMask(Mask):
    def __init__(self, extensions):
        if isinstance(extensions, str):
            extensions = {extensions}
        self.extensions = set(extensions)

    def __call__(self, element, **kwargs):
        return element.extension in self.extensions

    def __repr__(self):
        return f"chord_extension_in({self.extensions})"


class TonalityDegreeInMask(Mask):

    def __init__(self, degrees):
        if isinstance(degrees, int):
            degrees = {degrees}
        self.degrees = set(degrees)

    def __call__(self, element, **kwargs):
        return element.tonality.degree in self.degrees

    def __repr__(self):
        return f"tonality_degree_in({self.degrees})"




class FuncMask(Mask):

    def __init__(self, f):
        self.f = f

    def __call__(self, element, **kwargs):
        return self.f(element, **kwargs)

    def __repr__(self):
        return 'funcMask'


class HasAtLeastMask(Mask):
    def __init__(self, tags):
        if isinstance(tags, str):
            tags = {tags}
        self.tags = set(tags)

    def __call__(self, element, **kwargs):
        return len(element.tags.intersection(self.tags)) > 0

    def __repr__(self):
        return f"has_at_least({self.tags})"


class NotMask(Mask):
    def __init__(self, other):
        self.other = other

    def __call__(self, element, **kwargs):
        return not self.other(element, **kwargs)

    def __repr__(self):
        return f"~ ({self.other})"


class HasMask(Mask):
    def __init__(self, tags):
        if isinstance(tags, str):
            tags = {tags}
        self.tags = set(tags)

    def __call__(self, element, **kwargs):
        return element.tags.issuperset(self.tags)

    def __repr__(self):
        return f"has({self.tags})"

    def __invert__(self):
        return NotMask(self)


class AndMask(Mask):
    def __init__(self, terms):
        self.terms = terms

    def child(self, element, **kwargs):
        return self.__class__([t.child(element, **kwargs) for t in self.terms])

    def __call__(self, element, **kwargs):
        return all((t(element, **kwargs) for t in self.terms))

    def __repr__(self):
        return " & ".join([f"({term})" for term in self.terms])

    def __invert__(self):
        return OrMask([~m for m in self.terms])


class OrMask(Mask):
    def __init__(self, terms):
        self.terms = terms

    def child(self, element, **kwargs):
        return self.__class__([t.child(element, **kwargs) for t in self.terms])

    def __call__(self, element, **kwargs):
        return any((t(element, **kwargs) for t in self.terms))

    def __repr__(self):
        return " | ".join([f"({term})" for term in self.terms])

    def __invert__(self):
        return AndMask([~m for m in self.terms])


class BoolMask(Mask):

    def __init__(self, bool):
        self.bool = bool

    def __repr__(self):
        return f"{str(self.bool).lower()}"

    def __call__(self, element, **kwargs):
        return self.bool

    def __invert__(self):
        return BoolMask(not self.bool)


class FalseMask(Mask):

    def __call__(self, element, **kwargs):
        return False

    def __repr__(self):
        return "false"

    def __invert__(self):
        return BoolMask(True)


class TrueMask(Mask):

    def __call__(self, element, **kwargs):
        return True

    def __repr__(self):
        return "true"

    def __invert__(self):
        return BoolMask(False)


class GtMask(Mask):

    def __init__(self, terms):
        self.terms = terms

    def child(self, element, **kwargs):
        return self if self.terms[0](element, **kwargs) else BoolMask(self(element, **kwargs))

    def __call__(self, element, **kwargs):
        return self.terms[0](element, **kwargs) or self.terms[1](element, **kwargs)

    def __repr__(self):
        return " > ".join([f"({term})" for term in self.terms])

    def __invert__(self):
        return GtMask([self.terms[0], ~self.terms[1]])


class TypeMask(Mask):

    def __gt__(self, other):
        return GtMask([self, other])


class ScoreMask(TypeMask):

    def __call__(self, element, **kwargs):
        from musiclang import Score
        return not isinstance(element, Score)

    def __repr__(self):
        return "Score"

    def __invert__(self):
        return self


class MelodyMask(TypeMask):

    def __call__(self, element, **kwargs):
        from musiclang import Melody
        return not isinstance(element, Melody)

    def __repr__(self):
        return "Melody"

    def __invert__(self):
        return self


class ChordMask(TypeMask):
    def __call__(self, element, **kwargs):
        from musiclang import Chord
        return not isinstance(element, Chord)

    def __repr__(self):
        return "Chord"

    def __invert__(self):
        return self


class NoteMask(TypeMask):
    def __call__(self, element, **kwargs):
        from musiclang import Note
        return not isinstance(element, Note)

    def __repr__(self):
        return "Note"

    def __invert__(self):
        return self


